__all__ = ('GuildPreview', )

from ...bases import DiscordEntity, IconSlot
from ...http import urls as module_urls
from ...utils import DATETIME_FORMAT_CODE

from .fields import (
    parse_approximate_online_count, parse_approximate_user_count, parse_description, parse_emojis, parse_features,
    parse_id, parse_name, parse_stickers, put_approximate_online_count_into, put_approximate_user_count_into,
    put_description_into, put_emojis_into, put_features_into, put_id_into, put_name_into, put_stickers_into,
    validate_approximate_online_count, validate_approximate_user_count, validate_description, validate_emojis,
    validate_features, validate_id, validate_name, validate_stickers
)


class GuildPreview(DiscordEntity):
    """
    A preview of a public guild.
    
    Attributes
    ----------
    approximate_online_count : `int`
        Approximate amount of online users at the represented guild.
    approximate_user_count : `int`
        Approximate amount of users at the represented guild.
    description : `None`, `str`
        Description of the guild.
    discovery_splash_hash : `int`
        The guild's discovery splash's hash in `uint128`.
    discovery_splash_type : ``IconType``
        The guild discovery splash's type.
    emojis : `dict` of (`int`, ``Emoji``) items
        The emojis of the guild stored in `emoji_id` - `emoji` relation.
    features : `None`, `tuple` of ``GuildFeature``
        The guild's features.
    icon_hash : `int`
        The guild's icon's hash in `uint128`.
    icon_type : ``IconType``
        The guild's icon's type.
    id : `int`
        The represented guild's identifier.
    invite_splash_hash : `int`
        The guild's invite splash's hash in `uint128`.
    invite_splash_type : ``IconType``
        the guild's invite splash's type.
    stickers : `dict` of (`int`, ``Sticker``) items
        The stickers of the guild stored in `sticker_id` - `sticker` relation.
    name : `str`
        The name of the guild.
    """
    __slots__ = (
        'approximate_online_count', 'approximate_user_count','description', 'emojis', 'features', 'name', 'stickers'
    )
    
    icon = IconSlot(
        'icon',
        'icon',
        module_urls.guild_icon_url,
        module_urls.guild_icon_url_as,
    )
    
    invite_splash = IconSlot(
        'invite_splash',
        'splash',
        module_urls.guild_invite_splash_url,
        module_urls.guild_invite_splash_url_as,
    )
    
    discovery_splash = IconSlot(
        'discovery_splash',
        'discovery_splash',
        module_urls.guild_discovery_splash_url,
        module_urls.guild_discovery_splash_url_as,
    )
    
    
    def __new__(
        cls,
        *,
        approximate_online_count = ...,
        approximate_user_count = ...,
        description = ...,
        discovery_splash = ...,
        emojis = ...,
        features = ...,
        guild_id = ...,
        icon = ...,
        invite_splash = ...,
        stickers = ...,
        name = ...,
    ):
        """
        Creates a new guild preview with the given fields.
        
        Parameters
        ----------
        approximate_online_count : `int`, Optional (Keyword only)
            Approximate amount of online users at the represented guild.
        approximate_user_count : `int`, Optional (Keyword only)
            Approximate amount of users at the represented guild.
        description : `None`, `str`, Optional (Keyword only)
            Description of the guild.
        discovery_splash : `None`, ``Icon``, `str`, Optional (Keyword only)
            The guild's discovery splash.
        emojis : `None`, `iterable` of ``Emoji``, Optional (Keyword only)
            The emojis of the guild.
        features : `None`, `iterable` of ``GuildFeature``, Optional (Keyword only)
            The guild's features.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The represented guild's identifier.
        icon : `None`, ``Icon``, `str`, Optional (Keyword only)
            The guild's icon.
        invite_splash : `None`, ``Icon``, `str`, Optional (Keyword only)
            The guild's invite splash.
        stickers : `None`, `iterable` of ``Sticker``, Optional (Keyword only)
            The stickers of the guild.
        name : `str`, Optional (Keyword only)
            The name of the guild.
        
        Raises
        ------
        TypeError
            - Parameter with incorrect type given.
        ValueError
            - Parameter with incorrect value given.
        """
        # approximate_online_count
        if approximate_online_count is ...:
            approximate_online_count = 0
        else:
            approximate_online_count = validate_approximate_online_count(approximate_online_count)
        
        # approximate_user_count
        if approximate_user_count is ...:
            approximate_user_count = 0
        else:
            approximate_user_count = validate_approximate_user_count(approximate_user_count)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # discovery_splash
        if discovery_splash is ...:
            discovery_splash = None
        else:
            discovery_splash = cls.discovery_splash.validate_icon(discovery_splash)
        
        # emojis
        if emojis is ...:
            emojis = {}
        else:
            emojis = validate_emojis(emojis)
        
        # features
        if features is ...:
            features = None
        else:
            features = validate_features(features)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_id(guild_id)
        
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon)
        
        # invite_splash
        if invite_splash is ...:
            invite_splash = None
        else:
            invite_splash = cls.invite_splash.validate_icon(invite_splash)
    
        # stickers
        if stickers is ...:
            stickers = {}
        else:
            stickers = validate_stickers(stickers)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # Construct
        self = object.__new__(cls)
        self.approximate_online_count = approximate_online_count
        self.approximate_user_count = approximate_user_count
        self.description = description
        self.discovery_splash = discovery_splash
        self.emojis = emojis
        self.features = features
        self.icon = icon
        self.id = guild_id
        self.invite_splash = invite_splash
        self.stickers = stickers
        self.name = name
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a guild preview from the requested guild preview data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received guild preview data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        guild_id = parse_id(data)
        
        self = object.__new__(cls)
        self.approximate_online_count = parse_approximate_online_count(data)
        self.approximate_user_count = parse_approximate_user_count(data)
        self.description = parse_description(data)
        self._set_discovery_splash(data)
        self.emojis = parse_emojis(data, {}, guild_id)
        self.features = parse_features(data)
        self._set_icon(data)
        self.id = guild_id
        self._set_invite_splash(data)
        self.stickers = parse_stickers(data, {})
        self.name = parse_name(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the guild preview to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_approximate_online_count_into(self.approximate_online_count, data, defaults)
        put_approximate_user_count_into(self.approximate_user_count, data, defaults)
        put_description_into(self.description, data, defaults)
        type(self).discovery_splash.put_into(self.discovery_splash, data, defaults)
        put_emojis_into(self.emojis, data, defaults)
        put_features_into(self.features, data, defaults)
        type(self).icon.put_into(self.icon, data, defaults)
        put_id_into(self.id, data, defaults)
        type(self).invite_splash.put_into(self.invite_splash, data, defaults)
        put_stickers_into(self.stickers, data, defaults)
        put_name_into(self.name, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the guild preview's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        sticker_id = self.id
        if sticker_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __format__(self, code):
        """
        Formats the guild preview in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        channel : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import Client, KOKORO
        >>> TOKEN = 'a token goes here'
        >>> client = Client(TOKEN)
        >>> guild_id = 302094807046684672
        >>> guild_preview = KOKORO.run(client.guild_preview_get(guild_id))
        >>> guild_preview
        <GuildPreview id = 302094807046684672, name = 'MINECRAFT'>
        >>> # no code stands for `guild_preview.name`.
        >>> f'{guild_preview}'
        'MINECRAFT'
        >>> # 'c' stands for created at.
        >>> f'{guild_preview:c}'
        '2017.04.13-14:56:54'
        ```
        """
        if not code:
            return self.name
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}.'
        )
    
    
    def __eq__(self, other):
        """Returns whether the two guild previews are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two guild previews are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return (not self._is_equal_same_type(other))
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two guild previews are equal. Type of other must be same as type of self.
        
        Parameters
        ----------
        other : `<instance<type<self>>`
            The other guild preview to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        # approximate_online_count
        if self.approximate_online_count != other.approximate_online_count:
            return False
        
        # approximate_user_count
        if self.approximate_user_count != other.approximate_user_count:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # discovery_splash_hash
        if self.discovery_splash_hash != other.discovery_splash_hash:
            return False
        
        # discovery_splash_type
        if self.discovery_splash_type != other.discovery_splash_type:
            return False
        
        # emojis
        if self.emojis != other.emojis:
            return False
        
        # features
        if self.features != other.features:
            return False
        
        # icon_hash
        if self.icon_hash != other.icon_hash:
            return False
        
        # icon_type
        if self.icon_type != other.icon_type:
            return False
        
        # id
        if self.id != other.id:
            return False
        
        # invite_splash_hash
        if self.invite_splash_hash != other.invite_splash_hash:
            return False
        
        # invite_splash_type
        if self.invite_splash_type != other.invite_splash_type:
            return False
        
        # stickers
        if self.stickers != other.stickers:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the guild preview."""
        hash_value = 0
        
        # approximate_online_count
        hash_value ^= self.approximate_online_count
        
        # approximate_user_count
        hash_value ^= self.approximate_user_count << 12
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # discovery_splash
        hash_value ^= hash(self.discovery_splash)
        
        # emojis
        emojis = self.emojis
        hash_value ^= len(emojis) << 1
        for emoji in emojis.values():
            hash_value ^= hash(emoji)
        
        # features
        features = self.features
        hash_value ^= len(features) << 5
        for feature in features:
            hash_value ^= hash(feature)
        
        # icon
        hash_value ^= hash(self.icon)
        
        # id
        hash_value ^= self.id
        
        # invite_splash
        hash_value ^= hash(self.invite_splash)
        
        # stickers
        stickers = self.stickers
        hash_value ^= len(stickers) << 9
        for sticker in stickers.values():
            hash_value ^= hash(sticker)
        
        # name
        name = self.name
        if (description is None) or (description != name):
            hash_value ^= hash(name)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the guild preview.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.approximate_online_count = self.approximate_online_count
        new.approximate_user_count = self.approximate_user_count
        new.description = self.description
        new.discovery_splash_hash = self.discovery_splash_hash
        new.discovery_splash_type = self.discovery_splash_type
        new.emojis = self.emojis.copy()
        features = self.features
        if (features is not None):
            features = (*features,)
        new.features = features
        new.icon_hash = self.icon_hash
        new.icon_type = self.icon_type
        new.id = self.id
        new.invite_splash_hash = self.invite_splash_hash
        new.invite_splash_type = self.invite_splash_type
        new.stickers = self.stickers.copy()
        new.name = self.name
        return new
    
    
    def copy_with(
        self,
        *,
        approximate_online_count = ...,
        approximate_user_count = ...,
        description = ...,
        discovery_splash = ...,
        emojis = ...,
        features = ...,
        guild_id = ...,
        icon = ...,
        invite_splash = ...,
        stickers = ...,
        name = ...,
    ):
        """
        Copies the guild preview with the given fields.
        
        Parameters
        ----------
        approximate_online_count : `int`, Optional (Keyword only)
            Approximate amount of online users at the represented guild.
        approximate_user_count : `int`, Optional (Keyword only)
            Approximate amount of users at the represented guild.
        description : `None`, `str`, Optional (Keyword only)
            Description of the guild.
        discovery_splash : `None`, ``Icon``, `str`, Optional (Keyword only)
            The guild's discovery splash.
        emojis : `None`, `iterable` of ``Emoji``, Optional (Keyword only)
            The emojis of the guild.
        features : `None`, `iterable` of ``GuildFeature``, Optional (Keyword only)
            The guild's features.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The represented guild's identifier.
        icon : `None`, ``Icon``, `str`, Optional (Keyword only)
            The guild's icon.
        invite_splash : `None`, ``Icon``, `str`, Optional (Keyword only)
            The guild's invite splash.
        stickers : `None`, `iterable` of ``Sticker``, Optional (Keyword only)
            The stickers of the guild.
        name : `str`, Optional (Keyword only)
            The name of the guild.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Parameter with incorrect type given.
        ValueError
            - Parameter with incorrect value given.
        """
        # approximate_online_count
        if approximate_online_count is ...:
            approximate_online_count = self.approximate_online_count
        else:
            approximate_online_count = validate_approximate_online_count(approximate_online_count)
        
        # approximate_user_count
        if approximate_user_count is ...:
            approximate_user_count = self.approximate_user_count
        else:
            approximate_user_count = validate_approximate_user_count(approximate_user_count)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # discovery_splash
        if discovery_splash is ...:
            discovery_splash = self.discovery_splash
        else:
            discovery_splash = type(self).discovery_splash.validate_icon(discovery_splash)
        
        # emojis
        if emojis is ...:
            emojis = self.emojis.copy()
        else:
            emojis = validate_emojis(emojis)
        
        # features
        if features is ...:
            features = self.features
            if (features is not None):
                features = (*features,)
        else:
            features = validate_features(features)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.id
        else:
            guild_id = validate_id(guild_id)
        
        # icon
        if icon is ...:
            icon = self.icon
        else:
            icon = type(self).icon.validate_icon(icon)
        
        # invite_splash
        if invite_splash is ...:
            invite_splash = self.invite_splash
        else:
            invite_splash = type(self).invite_splash.validate_icon(invite_splash)
    
        # stickers
        if stickers is ...:
            stickers = self.stickers.copy()
        else:
            stickers = validate_stickers(stickers)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # Construct
        new = object.__new__(type(self))
        new.approximate_online_count = approximate_online_count
        new.approximate_user_count = approximate_user_count
        new.description = description
        new.discovery_splash = discovery_splash
        new.emojis = emojis
        new.features = features
        new.icon = icon
        new.id = guild_id
        new.invite_splash = invite_splash
        new.stickers = stickers
        new.name = name
        return new
    
    
    def iter_features(self):
        """
        Iterates over the features of the represented guild.
        
        This method is an iterable generator.
        
        Yields
        ------
        feature : ``GuildFeature``
        """
        features = self.features
        if (features is not None):
            yield from features
    
    
    def has_feature(self, feature):
        """
        Returns whether the represented guild has the give feature.
        
        Parameters
        ----------
        feature : ``GuildFeature``
            The feature to look for.
        
        Returns
        -------
        has_feature : `bool`
        """
        features = self.features
        if features is None:
            return False
        
        return feature in features



__all__ = ()

from scarletio import CallableAnalyzer, include_with_callback

from .utils import datetime_to_timestamp


def _has_entity_include_internals_parameter(entity_type):
    """
    Returns whether the given entity's `.to_data` method has `include_internals` parameter.
    
    Parameters
    ----------
    entity_type : `type` with `{to_data}`
        The entity type to inspect.
    
    Returns
    -------
    has_include_internal_parameter : `bool`
    """
    analyzer = CallableAnalyzer(entity_type.to_data, as_method = True)
    
    for parameter in analyzer.parameters:
        if parameter.name == 'include_internals':
            has_include_internals_parameter = True
            break
    
    else:
        has_include_internals_parameter = False
    
    return has_include_internals_parameter


def entity_id_optional_putter_factory(field_key):
    """
    Returns an optional entity id putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(entity_id, data, defaults):
        """
        Puts the `entity_id` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        entity_id : `int`
            An entity's identifier.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults or entity_id:
            if entity_id:
                raw_entity_id = str(entity_id)
            else:
                raw_entity_id = None
            data[field_key] = raw_entity_id
        
        return data
    
    return putter


def entity_id_putter_factory(field_key):
    """
    Returns an entity id putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(entity_id, data, defaults):
        """
        Puts the `entity_id` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        entity_id : `int`
            An entity's identifier.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if entity_id:
            raw_entity_id = str(entity_id)
        else:
            raw_entity_id = None
        data[field_key] = raw_entity_id
    
        return data
    
    return putter


def optional_entity_id_array_optional_putter_factory(field_key):
    """
    Returns a nullable  entity id array optional putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(entity_id_array, data, defaults):
        """
        Puts the `entity_id_array` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        entity_id_array : `None`, `tuple` of `int`
            An entity's identifier.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
            
        if defaults or (entity_id_array is not None):
            if entity_id_array is None:
                entity_id_array = []
            else:
                entity_id_array = [str(entity_id) for entity_id in entity_id_array]
            
            data[field_key] = entity_id_array
        
        return data
    
    return putter


def nullable_entity_id_array_putter_factory(field_key):
    """
    Returns a nullable entity id array putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(entity_id_array, data, defaults):
        """
        Puts the `entity_id_array` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        entity_id_array : `None`, `tuple` of `int`
            An entity's identifier.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if entity_id_array is None:
            entity_id_array = []
        else:
            entity_id_array = [str(entity_id) for entity_id in entity_id_array]
        
        data[field_key] = entity_id_array
        
        return data
    
    return putter


def preinstanced_putter_factory(field_key):
    """
    Returns a preinstanced putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(preinstanced, data, defaults):
        """
        Puts the `preinstanced` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        preinstanced : ``PreinstancedBase``
            The preinstanced object.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        data[field_key] = preinstanced.value
        
        return data
    
    return putter


def preinstanced_optional_putter_factory(field_key, default):
    """
    Returns an optional preinstanced putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default : `object`
        Default value to exclude when default values are not required.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(preinstanced, data, defaults):
        """
        Puts the `preinstanced` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        preinstanced : ``PreinstancedBase``
            The preinstanced object.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal default
        nonlocal field_key
        
        if defaults or (preinstanced is not default):
            data[field_key] = preinstanced.value
        
        return data
    
    return putter


def preinstanced_array_putter_factory(field_key):
    """
    Returns a new preinstanced array putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(preinstanced_array, data, defaults):
        """
        Puts the `preinstanced` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        preinstanced_array :`None`, `tuple` of  ``PreinstancedBase``
            The preinstanced object.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if preinstanced_array is None:
            field_value = []
        else:
            field_value = [preinstanced.value for preinstanced in preinstanced_array]
        data[field_key] = field_value
        
        return data
    
    return putter


def int_putter_factory(field_key):
    """
    Returns an `int` putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    return field_putter_factory(field_key)


def field_putter_factory(field_key):
    """
    Returns a field putter which 1:1 puts the given value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given object into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `object`
            Object field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        data[field_key] = field_value
        
        return data
    
    return putter


def nullable_field_optional_putter_factory(field_key):
    """
    Returns a nullable optional field putter which 1:1 puts the given value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given object into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `object`
            Object field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults or (field_value is not None):
            data[field_key] = field_value
        
        return data
    
    return putter


def field_optional_putter_factory(field_key, default_value):
    """
    Returns an optional field putter which 1:1 puts the given value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `object`
        The default value to ignore if defaults are not required.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given object into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `object`
            Object field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal default_value
        nonlocal field_key
        
        if defaults or (field_value != default_value):
            data[field_key] = field_value
        
        return data
    
    return putter


def bool_optional_putter_factory(field_key, default_value):
    """
    Returns an optional `bool` putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `bool`
        The default value to ignore if defaults are not required.
    
    Returns
    -------
    putter : `FunctionType`
    """
    return field_optional_putter_factory(field_key, default_value)


def float_optional_putter_factory(field_key, default_value):
    """
    Returns an optional `float` putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `float`
        The default value to ignore if defaults are not required.
    
    Returns
    -------
    putter : `FunctionType`
    """
    return field_optional_putter_factory(field_key, default_value)


def int_optional_putter_factory(field_key, default_value):
    """
    Returns an optional `int` putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `int`
        The default value to ignore if defaults are not required.
    
    Returns
    -------
    putter : `FunctionType`
    """
    return field_optional_putter_factory(field_key, default_value)


def int_optional_postprocess_putter_factory(field_key, default_value, postprocessor):
    """
    Returns an optional `int` putter with a postprocessor function which is called if the value should be put in the
    payload.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `int`
        The default value to ignore if defaults are not required.
    postprocessor : `callable`
        Postprocessor to apply before putting the value into the the `data` object.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `int` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `int`
            Integer field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        nonlocal default_value
        nonlocal postprocessor
        
        if defaults or (field_value != default_value):
            data[field_key] = postprocessor(field_value)
        
        return data
    
    return putter


def nullable_int_optional_putter_factory(field_key, default_value):
    """
    Returns an nullable optional `int` putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `int`
        The default value to ignore if defaults are not required.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `int` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `int`
            Integer field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        nonlocal default_value
        
        if defaults or (field_value != default_value):
            if field_value == default_value:
                field_value = None
            
            data[field_key] = field_value
        
        return data
    
    return putter


def force_bool_putter_factory(field_key):
    """
    Returns a forced `bool` putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `bool` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `bool`
            Boolean field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        data[field_key] = field_value
        
        return data
    
    return putter


def negated_bool_optional_putter_factory(field_key, default_value):
    """
    Returns an negated optional `bool` putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `bool`
        The default value to ignore if defaults are not required.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Negates the given `bool` and puts it into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `bool`
            Boolean field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        nonlocal default_value
        
        if defaults or (field_value != default_value):
            data[field_key] = not field_value
        
        return data
    
    return putter


def nullable_date_time_optional_putter_factory(field_key):
    """
    Returns a new nullable date time putter.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `DateTime` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `DateTime`
            Date time field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults or (field_value is not None):
            if field_value is None:
                timestamp = None
            else:
                timestamp = datetime_to_timestamp(field_value)
            data[field_key] = timestamp
        
        return data
    
    return putter


def force_string_putter_factory(field_key):
    """
    Returns a new string putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `string` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `string`
            String field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        data[field_key] = field_value
        
        return data
    
    return putter


def nullable_string_putter_factory(field_key):
    """
    Returns a nullable string putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `string` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `None`, `string`
            String field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if field_value is None:
            field_value = ''
        
        data[field_key] = field_value
        
        return data
    
    return putter


def nullable_string_optional_putter_factory(field_key):
    """
    Returns a nullable & optional string putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `string` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `None`, `string`
            String field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults:
            if field_value is None:
                field_value = ''
        
        else:
            if field_value is None:
                return data
        
        data[field_key] = field_value
        return data
    
    return putter


def url_optional_putter_factory(field_key):
    """
    Returns an optional `url` putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `string` into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `None`, `string`
            String field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults or (field_value is not None):
            data[field_key] = field_value
        
        return data
    
    return putter


def nullable_string_array_optional_putter_factory(field_key):
    """
    Returns an nullable optional string putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `string` array into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `None`, `tuple` of `str`
            String field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults or (field_value is not None):
            if field_value is None:
                string_array = []
            else:
                string_array = [*field_value]
            
            data[field_key] = string_array
        
        return data
    
    return putter


def nullable_entity_array_putter_factory(
    field_key, field_type, *, can_include_internals = ..., force_include_internals = False, include = None
):
    """
    Returns a new nullable entity array putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    field_type : `type`
        The field's type.
    
    can_include_internals : `bool`, Optional (Keyword only)
        Whether the `field_type.to_data` implements the `include_internals` parameter.
    
    force_include_internals : `bool`, Optional (Keyword only)
        Whether `include_internals` should be passed as `True` always.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    putter : `FunctionType`
    """
    if force_include_internals:
        def putter(entity_array, data, defaults):
            """
            Puts the given entity array into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_array : `None`, `tuple` of `object`
                Entity array.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
                
            if entity_array is None:
                entity_data_array = []
            else:
                entity_data_array = [
                    entity.to_data(defaults = defaults, include_internals = True) for entity in entity_array
                ]
            
            data[field_key] = entity_data_array
            
            return data
        
    elif (
        ((can_include_internals is not ...) and can_include_internals) or
        ((field_type is not NotImplemented) and _has_entity_include_internals_parameter(field_type))
    ):
        def putter(entity_array, data, defaults, *, include_internals = False):
            """
            Puts the given entity array into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_array : `None`, `tuple` of `object`
                Entity array.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            include_internals : `bool` = `False`, Optional (Keyword only)
                Whether internal fields should be included.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
                
            if entity_array is None:
                entity_data_array = []
            else:
                entity_data_array = [
                    entity.to_data(defaults = defaults, include_internals = include_internals)
                    for entity in entity_array
                ]
            
            data[field_key] = entity_data_array
            
            return data
    
    else:
        def putter(entity_array, data, defaults):
            """
            Puts the given entity array into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_array : `None`, `tuple` of `object`
                Entity array.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
                
            if entity_array is None:
                entity_data_array = []
            else:
                entity_data_array = [entity.to_data(defaults = defaults) for entity in entity_array]
            
            data[field_key] = entity_data_array
            
            return data
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_field_type(value):
            nonlocal field_type
            field_type = value
    
    
    return putter


def _iter_nullable_dictionary_values(dictionary):
    """
    Iterates over a nullable dictionary's values.
    
    Parameters
    ----------
    dictionary : `dict<KeyType, ValueType>
        Dictionary to iterate over.
    
    Yields
    ------
    value : `ValueType`
    """
    if (dictionary is not None):
        yield from dictionary.values()


def entity_dictionary_putter_factory(
    field_key, entity_type, *, can_include_internals = ..., force_include_internals = False, include = None
):
    """
    Returns an entity dictionary putter factory.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    field_type : `type`
        The field's type.
    
    can_include_internals : `bool`, Optional (Keyword only)
        Whether the `field_type.to_data` implements the `include_internals` parameter.
    
    force_include_internals : `bool`, Optional (Keyword only)
        Whether `include_internals` should be passed as `True` always.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    putter : `FunctionType`
    """
    if force_include_internals:
        def putter(entity_dictionary, data, defaults):
            """
            Puts the given entity dictionary into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_dictionary : `None`, `dict` of (`int`, `object`) items.
                Entity dictionary.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            data[field_key] = [
                entity.to_data(defaults = defaults, include_internals = True)
                for entity in _iter_nullable_dictionary_values(entity_dictionary)
            ]
            return data
    elif (
        ((can_include_internals is not ...) and can_include_internals) or
        ((entity_type is not NotImplemented) and _has_entity_include_internals_parameter(entity_type))
    ):
        def putter(entity_dictionary, data, defaults, *, include_internals = False):
            """
            Puts the given entity dictionary into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_dictionary : `None`, `dict` of (`int`, `object`) items.
                Entity dictionary.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            include_internals : `bool` = `False`, Optional (Keyword only)
                Whether internal fields should be included.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            data[field_key] = [
                entity.to_data(defaults = defaults, include_internals = include_internals)
                for entity in _iter_nullable_dictionary_values(entity_dictionary)
            ]
            return data
    else:
        def putter(entity_dictionary, data, defaults):
            """
            Puts the given entity dictionary into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_dictionary : `None`, `dict` of (`int`, `object`) items.
                Entity dictionary.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            data[field_key] = [
                entity.to_data(defaults = defaults)
                for entity in _iter_nullable_dictionary_values(entity_dictionary)
            ]
            return data
        
    if (include is not None):
        @include_with_callback(include)
        def include_field_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    return putter


def nullable_entity_array_optional_putter_factory(
    field_key, entity_type, *, can_include_internals = ..., include = None
):
    """
    Returns a nullable entity array putter.
    
    Parameters
    ---------
    field_key : `str`
        The field's key used in payload.
    
    field_type : `type`
        The field's type.
    
    can_include_internals : `bool`, Optional (Keyword only)
        Whether the `field_type.to_data` implements the `include_internals` parameter.
        
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    putter : `FunctionType`
    """
    if (
        ((can_include_internals is not ...) and can_include_internals) or
        ((entity_type is not NotImplemented) and _has_entity_include_internals_parameter(entity_type))
    ):
        def putter(entity_array, data, defaults, *, include_internals = False):
            """
            Puts the given entity array into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_array : `None`, `tuple` of `object`
                Entity array.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            include_internals : `bool` = `False`, Optional (Keyword only)
                Whether internal fields should be included.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            if defaults or (entity_array is not None):
                if entity_array is None:
                    entity_data_array = []
                else:
                    entity_data_array = [
                        entity.to_data(defaults = defaults, include_internals = include_internals)
                        for entity in entity_array
                    ]
                
                data[field_key] = entity_data_array
            
            return data
    
    else:
        def putter(entity_array, data, defaults):
            """
            Puts the given entity array into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity_array : `None`, `tuple` of `object`
                Entity array.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            if defaults or (entity_array is not None):
                if entity_array is None:
                    entity_data_array = []
                else:
                    entity_data_array = [
                        entity.to_data(defaults = defaults)
                        for entity in entity_array
                    ]
                
                data[field_key] = entity_data_array
            
            return data

    
    if (include is not None):
        @include_with_callback(include)
        def include_field_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    return putter


def nullable_object_array_optional_putter_factory(field_key):
    """
    Returns a nullable object array putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(entity_array, data, defaults):
        """
        Puts the given entity array into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        entity_array : `None`, `tuple` of `object`
            Entity array.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults or (entity_array is not None):
            if entity_array is None:
                entity_data_array = []
            else:
                entity_data_array = [entity.to_data(defaults = defaults) for entity in entity_array]
            
            data[field_key] = entity_data_array
        
        return data
    
    return putter


def nullable_entity_putter_factory(field_key, entity_type):
    """
    Returns a new nullable entity putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    entity_type : `object` with `{to_data}`
        The expected entity type.
    
    Returns
    -------
    putter : `FunctionType`
    """
    return default_entity_putter_factory(field_key, entity_type, None)


def default_entity_putter_factory(field_key, entity_type, default, *, force_include_internals = False):
    """
    Returns a new defaulted entity putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    entity_type : `object` with `{to_data}`
        The expected entity type.
    
    default : `object`
        The default value to handle as an unique case.
    
    force_include_internals : `bool`, Optional (Keyword only)
        Whether `include_internals` should be passed as `True` always.
    
    Returns
    -------
    putter : `FunctionType`
    """
    if force_include_internals:
        def putter(entity, data, defaults):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal default
            nonlocal field_key
            
            if entity is default:
                entity_data = None
            else:
                entity_data = entity.to_data(defaults = defaults, include_internals = True)
            
            data[field_key] = entity_data
            
            return data
    
    elif _has_entity_include_internals_parameter(entity_type):
        def putter(entity, data, defaults, *, include_internals = False):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            include_internals : `bool` = `False`, Optional (Keyword only)
                Whether internal fields should be included.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal default
            nonlocal field_key
            
            if entity is default:
                entity_data = None
            else:
                entity_data = entity.to_data(defaults = defaults, include_internals = include_internals)
            
            data[field_key] = entity_data
            
            return data
    
    else:
        def putter(entity, data, defaults):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal default
            nonlocal field_key
            
            if entity is default:
                entity_data = None
            else:
                entity_data = entity.to_data(defaults = defaults)
            
            data[field_key] = entity_data
            
            return data
    
    return putter


def nullable_entity_optional_putter_factory(
    field_key, entity_type, *, can_include_internals = ..., force_include_internals = False, include = None
):
    """
    Returns a nullable optional entity putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    entity_type : `object` with `{to_data}`
        The expected entity type.
    
    default : `object`
        The default value to handle as an unique case.
    
    can_include_internals : `bool`, Optional (Keyword only)
        Whether the `field_type.to_data` implements the `include_internals` parameter.
        
    force_include_internals : `bool`, Optional (Keyword only)
        Whether `include_internals` should be passed as `True` always.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The object's name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    putter : `FunctionType`
    """
    if force_include_internals:
        def putter(entity, data, defaults):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            if defaults or (entity is not None):
                if entity is None:
                    entity_data = None
                else:
                    entity_data = entity.to_data(defaults = defaults, include_internals = True)
                
                data[field_key] = entity_data
            
            return data
    elif (
        ((can_include_internals is not ...) and can_include_internals) or
        ((entity_type is not NotImplemented) and _has_entity_include_internals_parameter(entity_type))
    ):
        def putter(entity, data, defaults, *, include_internals = False):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            include_internals : `bool` = `False`, Optional (Keyword only)
                Whether internal fields should be included.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            if defaults or (entity is not None):
                if entity is None:
                    entity_data = None
                else:
                    entity_data = entity.to_data(defaults = defaults, include_internals = include_internals)
                
                data[field_key] = entity_data
            
            return data
    
    else:
        def putter(entity, data, defaults):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            if defaults or (entity is not None):
                if entity is None:
                    entity_data = None
                else:
                    entity_data = entity.to_data(defaults = defaults)
                
                data[field_key] = entity_data
            
            return data
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_field_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    return putter


def entity_putter_factory(field_key, entity_type, *, force_include_internals = False):
    """
    Returns a new entity putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    entity_type : `object` with `{to_data}`
        The expected entity type.
        
    force_include_internals : `bool`, Optional (Keyword only)
        Whether `include_internals` should be passed as `True` always.
    
    
    Returns
    -------
    putter : `FunctionType`
    """
    if force_include_internals:
        def putter(entity, data, defaults):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            data[field_key] = entity.to_data(defaults = defaults, include_internals = True)
            
            return data
    
    elif _has_entity_include_internals_parameter(entity_type):
        def putter(entity, data, defaults, *, include_internals = False):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            include_internals : `bool` = `False`, Optional (Keyword only)
                Whether internal fields should be included.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            data[field_key] = entity.to_data(defaults = defaults, include_internals = include_internals)
            
            return data
    
    else:
        def putter(entity, data, defaults):
            """
            Puts the given entity into the given `data` json serializable object.
            
            > This function is generated.
            
            Parameters
            ----------
            entity : `object` with `{to_data}`
                Entity.
            data : `dict` of (`str`, `object`) items
                Json serializable dictionary.
            defaults : `bool`
                Whether default values should be included as well.
            
            Returns
            -------
            data : `dict` of (`str`, `object`) items
            """
            nonlocal field_key
            
            data[field_key] = entity.to_data(defaults = defaults)
            
            return data
    
    return putter



def nullable_functional_optional_putter_factory(field_key, function, *, include = None):
    """
    Returns a new nullable optional functional putter. If the given `field_value` is not `None`, it will call the
    function on field's value and put it into the received `data`.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    function : `FunctionType`
        The function to call to serialise the field value.
    include : `None`, `str` = `None`, Optional (Keyword only)
        The function's name to include `function` with. Should be used when `function` cannot be resolved initially.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `field_value` into the given `data` json serializable object. The `field_value` is processed by
        a function which is defined at the putter's creation.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `object`
            Field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal function
        nonlocal field_key
        
        if defaults or (field_value is not None):
            if field_value is not None:
                field_value = function(field_value)
            
            data[field_key] = field_value
        
        return data
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal function
            function = value
    
    
    return putter


def nullable_functional_array_optional_putter_factory(field_key, function, *, include = None):
    """
    Returns a nullable  array optional functional putter. If the given `field_value` is not `None`, it will call the
    function on field's value and put it into the received `data`.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    function : `FunctionType`
        The function to call to serialise the field value.
    include : `None`, `str` = `None`, Optional (Keyword only)
        The function's name to include `function` with. Should be used when `function` cannot be resolved initially.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(values, data, defaults):
        """
        Puts the given `field_value` into the given `data` json serializable object. The `field_value` is processed by
        a function which is defined at the putter's creation.
        
        > This function is generated.
        
        Parameters
        ----------
        values : `None`, `tuple` of `object`
            Field values.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal function
        nonlocal field_key
        
        if defaults or (values is not None):
            if values is None:
                field_value = []
            else:
                field_value = [function(value) for value in values]
            
            data[field_key] = field_value
        
        return data
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal function
            function = value
    
    
    return putter


def functional_putter_factory(field_key, function, *, include = None):
    """
    Returns a new functional putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    function : `FunctionType`
        The function to call to serialise the field value.
    include : `None`, `str` = `None`, Optional (Keyword only)
        The function's name to include `function` with. Should be used when `function` cannot be resolved initially.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(field_value, data, defaults):
        """
        Puts the given `field_value` into the given `data` json serializable object. The `field_value` is processed by
        a function which is defined at the putter's creation.
        
        > This function is generated.
        
        Parameters
        ----------
        field_value : `object`
            Field value.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal function
        nonlocal field_key
        
        data[field_key] = function(field_value)
        return data
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal function
            function = value
    
    
    return putter


def flag_putter_factory(field_key):
    """
    Returns a flag putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(flag, data, defaults):
        """
        Puts the given flags into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        flag : ``FlagBase``
            Flag instance.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        data[field_key] = int(flag)
        
        return data
    
    return putter


def flag_optional_putter_factory(field_key, default_value):
    """
    Returns a new defaulted flag putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    default_value : `object`
        The default value to handle as an unique case.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(flag, data, defaults):
        """
        Puts the given flags into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        flag : ``FlagBase``
            Flag instance.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        nonlocal default_value
        
        if defaults or (flag != default_value):
            data[field_key] = int(flag)
        
        return data
    
    return putter


def string_flag_optional_putter_factory(field_key, default_value):
    """
    Returns a defaulted string flag putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    default : `object`
        The default value to handle as an unique case.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(flag, data, defaults):
        """
        Puts the given flags into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        flag : ``FlagBase``
            Flag instance.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        nonlocal default_value
        
        if defaults or (flag != default_value):
            data[field_key] = format(flag, 'd')
        
        return data
    
    return putter


def nullable_flag_optional_putter_factory(field_key):
    """
    Returns a new nullable flag putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    default_value : `object`
        The default value to handle as an unique case.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(flag, data, defaults):
        """
        Puts the given flags into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        flag : ``FlagBase``
            Flag instance.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        if defaults or (flag is not None):
            if flag is not None:
                flag = int(flag)
            
            data[field_key] = flag
        
        return data
    
    return putter


def string_flag_putter_factory(field_key):
    """
    Returns a string flag putter.
    
    Returns
    -------
    field_key : `str`
        The field's key used in payload.
    default : `object`
        The default value to handle as an unique case.
    
    Returns
    -------
    putter : `FunctionType`
    """
    def putter(flag, data, defaults):
        """
        Puts the given flags into the given `data` json serializable object.
        
        > This function is generated.
        
        Parameters
        ----------
        flag : ``FlagBase``
            Flag instance.
        data : `dict` of (`str`, `object`) items
            Json serializable dictionary.
        defaults : `bool`
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        nonlocal field_key
        
        data[field_key] = format(flag, 'd')
        
        return data
    
    return putter



__all__ = ()

from scarletio import include

from ...application import Application
from ...bases import id_sort_key
from ...component import Component
from ...embed import Embed
from ...emoji import ReactionMapping, merge_update_reaction_mapping
from ...field_parsers import (
    bool_parser_factory, entity_id_array_parser_factory, entity_id_parser_factory, flag_parser_factory,
    nullable_date_time_parser_factory, nullable_entity_array_parser_factory, nullable_entity_parser_factory,
    nullable_functional_array_parser_factory, nullable_object_array_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, flag_optional_putter_factory,
    nullable_date_time_optional_putter_factory, nullable_entity_array_putter_factory,
    nullable_entity_optional_putter_factory, nullable_functional_array_optional_putter_factory,
    nullable_object_array_optional_putter_factory, nullable_string_optional_putter_factory,
    optional_entity_id_array_optional_putter_factory, preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator, entity_id_array_validator_factory, entity_id_validator_factory,
    flag_validator_factory, nullable_date_time_validator_factory, nullable_entity_array_validator_factory,
    nullable_entity_validator_factory, nullable_object_array_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)
from ...role import Role
from ...sticker import Sticker, create_partial_sticker_data, create_partial_sticker_from_partial_data
from ...user import ClientUserBase, User, UserBase, ZEROUSER
from ...webhook import WebhookBase, WebhookRepr, WebhookType, create_partial_webhook_from_id

from ..attachment import Attachment
from ..message_activity import MessageActivity
from ..message_application import MessageApplication
from ..message_call import MessageCall
from ..message_interaction import MessageInteraction
from ..message_role_subscription import MessageRoleSubscription

from .constants import CONTENT_LENGTH_MAX, NONCE_LENGTH_MAX
from .flags import MessageFlag
from .preinstanced import MessageType


Channel = include('Channel')
Message = include('Message')

# activity

parse_activity = nullable_entity_parser_factory('activity', MessageActivity)
put_activity_into = nullable_entity_optional_putter_factory('activity', MessageActivity)
validate_activity = nullable_entity_validator_factory('activity', MessageActivity)

# application

parse_application = nullable_entity_parser_factory('application', MessageApplication)
put_application_into = nullable_entity_optional_putter_factory('application', MessageApplication)
validate_application = nullable_entity_validator_factory('application', MessageApplication)

# application_id

parse_application_id = entity_id_parser_factory('application_id')


def put_application_id_into(application_id, data, defaults):
    """
    Puts the message's application's identifier into the given `data` json serializable object.
    
    Parameters
    ----------
    application_id : `int`
        Application identifier.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if application_id:
        raw_application_id = str(application_id)
        data['application_id'] = raw_application_id
        
        if data.get('webhook_id', None) is None:
            data['webhook_id'] = raw_application_id
    
    elif defaults:
        data['application_id'] = None
    
    return data

validate_application_id = entity_id_validator_factory('application_id', Application)

# attachments

parse_attachments = nullable_entity_array_parser_factory('attachments', Attachment)
put_attachments_into = nullable_entity_array_putter_factory('attachments', Attachment)
validate_attachments = nullable_entity_array_validator_factory('attachments', Attachment)

# author

def parse_author(data, guild_id = 0, channel_id = 0):
    """
    Parses the message's author from its data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message data.
    guild_id : `int` = `0`, Optional
        The guild's id where the message was created at.
    channel_id : `int` = `0`, Optional
        The channel's id where the message was created at.
    
    Returns
    -------
    author : ``UserBase``
    """
    author_data = data.get('author', None)
    
    webhook_id = data.get('webhook_id', None)
    if (webhook_id is not None):
        application_id = data.get('application_id', None)
        if ((application_id is None) or (webhook_id != application_id)):
            if data.get('message_reference', None) is None:
                webhook_type = WebhookType.bot
            else:
                webhook_type = WebhookType.server
            
            webhook_id = int(webhook_id)
            if author_data is None:
                return create_partial_webhook_from_id(webhook_id, '', webhook_type = webhook_type)
            
            return WebhookRepr.from_data(author_data, webhook_id, webhook_type, channel_id)
    
    if author_data is None:
        return ZEROUSER
    
    return User.from_data(author_data, data.get('member', None), guild_id)


def put_author_into(author, data, defaults, *, guild_id = 0):
    """
    Puts the message author's data into the given `data` json serializable object.
    
    Parameters
    ----------
    author : ``UserBase``
        Message author.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's id where the message was created at.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data['author'] =  author.to_data(defaults = defaults, include_internals = True)
    
    if guild_id:
        try:
            guild_profile = author.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            data['member'] = guild_profile.to_data(defaults = defaults, include_internals = True)
    
    # If the message as created by a webhook, put its into the data too.
    if isinstance(author, WebhookBase):
        data['webhook_id'] = str(author.id)
    elif defaults:
        data['webhook_id'] = None
    
    return data


validate_author = default_entity_validator('author', UserBase, default = ZEROUSER)

# call

parse_call = nullable_entity_parser_factory('call', MessageCall)
put_call_into = nullable_entity_optional_putter_factory('call', MessageCall)
validate_call = nullable_entity_validator_factory('call', MessageCall)

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id_into = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', NotImplemented, include = 'Channel')

# components

parse_components = nullable_object_array_parser_factory('components', Component)
put_components_into = nullable_object_array_optional_putter_factory('components')
validate_components = nullable_object_array_validator_factory('components', Component)

# content

parse_content = nullable_string_parser_factory('content')
put_content_into = nullable_string_optional_putter_factory('content')
validate_content = nullable_string_validator_factory('content', 0, CONTENT_LENGTH_MAX)

# embed

parse_embeds = nullable_object_array_parser_factory('embeds', Embed)
put_embeds_into = nullable_object_array_optional_putter_factory('embeds')
validate_embeds = nullable_object_array_validator_factory('embeds', Embed)

# edited_at

parse_edited_at = nullable_date_time_parser_factory('edited_timestamp')
put_edited_at_into = nullable_date_time_optional_putter_factory('edited_timestamp')
validate_edited_at = nullable_date_time_validator_factory('edited_at')

# flags

parse_flags = flag_parser_factory('flags', MessageFlag)
put_flags_into = flag_optional_putter_factory('flags', MessageFlag())
validate_flags = flag_validator_factory('flags', MessageFlag)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('message_id')

# interaction

def parse_interaction(data, guild_id = 0):
    """
    Parses the message's interaction from its data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message data.
    guild_id : `int` = `0`, Optional
        The guild's id where the message was created at.
    
    Returns
    -------
    interaction : `None`, ``MessageInteraction``
    """
    message_interaction_data = data.get('interaction', None)
    if (message_interaction_data is not None):
        return MessageInteraction.from_data(message_interaction_data, guild_id)


def put_interaction_into(interaction, data, defaults, *, guild_id = 0):
    """
    Puts the message interaction's data into the given `data` json serializable object.
    
    Parameters
    ----------
    interaction : `None`, ``MessageInteraction``
        Message interaction.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's id where the message was created at.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (interaction is not None):
        data['interaction'] = interaction.to_data(defaults = defaults, include_internals = True, guild_id = guild_id)
    elif defaults:
        data['interaction'] = None
    
    return data

validate_interaction = nullable_entity_validator_factory('interaction', MessageInteraction)

# mentioned_channels_cross_guild

parse_mentioned_channels_cross_guild = nullable_functional_array_parser_factory(
    'mention_channels',
    NotImplemented,
    include = 'create_partial_channel_from_data',
    sort_key = id_sort_key,
)
put_mentioned_channels_cross_guild_into = nullable_functional_array_optional_putter_factory(
    'mention_channels',
    NotImplemented,
    include = 'create_partial_channel_data',
)
validate_mentioned_channels_cross_guild = nullable_entity_array_validator_factory(
    'mentioned_channels_cross_guild', NotImplemented, include = 'Channel', sort_key = id_sort_key,
)

# mentioned_everyone

parse_mentioned_everyone = bool_parser_factory('mention_everyone', False)
put_mentioned_everyone_into = bool_optional_putter_factory('mention_everyone', False)
validate_mentioned_everyone = bool_validator_factory('mention_everyone', False)

# mentioned_role_ids

parse_mentioned_role_ids = entity_id_array_parser_factory('mention_roles')
put_mentioned_role_ids_into = optional_entity_id_array_optional_putter_factory('mention_roles')
validate_mentioned_role_ids = entity_id_array_validator_factory('mentioned_role_ids', Role)

# mentioned_users

def parse_mentioned_users(data, guild_id = 0):
    """
    Parses out the mentioned users from the given message data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message data.
    guild_id : `int` = `0`, Optional
        The guild's id where the message was created at.
    
    Returns
    -------
    mentioned_users : `None`, `tuple` of ``ClientUserBase``
    """
    user_mention_datas = data.get('mentions', None)
    if (user_mention_datas is None) or (not user_mention_datas):
        return
    
    return tuple(sorted(
        (
            User.from_data(user_mention_data, user_mention_data.get('member', None), guild_id)
            for user_mention_data in user_mention_datas
        ),
        key = id_sort_key,
    ))


def put_mentioned_users_into(mentioned_users, data, defaults, *, guild_id = 0):
    """
    Puts the message's mentioned users' data into the given `data` json serializable object.
    
    Parameters
    ----------
    mentioned_users : `None`, `tuple` of ``ClientUserBase``
        The mentioned users.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's id where the message was created at.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (mentioned_users is not None):
        user_mention_datas = []
        
        for user in mentioned_users:
            user_data = user.to_data(defaults = defaults, include_internals = True)
            
            if guild_id:
                try:
                    guild_profile = user.guild_profiles[guild_id]
                except KeyError:
                    pass
                else:
                    user_data['member'] = guild_profile.to_data(defaults = defaults, include_internals = True)
            
            user_mention_datas.append(user_data)
        
        data['mentions'] = user_mention_datas
    
    elif defaults:
        data['mentions'] = []
    
    return data


validate_mentioned_users = nullable_entity_array_validator_factory(
    'mentioned_users', ClientUserBase, sort_key = id_sort_key,
)

# message_id

def parse_message_id(data):
    """
    Parses out a message id from the given data. Not like `parse_id` it first checks for `message_id`
    and then for `id` keys.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message data.
    
    Returns
    -------
    message_id : `int`
    """
    try:
        message_id = data['message_id']
    except KeyError:
        pass
    else:
        if message_id is None:
            return 0
        
        return int(message_id)
    
    try:
        message_id = data['id']
    except KeyError:
        pass
    else:
        if message_id is None:
            return 0
        
        return int(message_id)
    
    return 0

put_message_id_into = entity_id_putter_factory('message_id')

# nonce

parse_nonce = nullable_string_parser_factory('nonce')
put_nonce_into = url_optional_putter_factory('nonce')
validate_nonce = nullable_string_validator_factory('nonce', 0, NONCE_LENGTH_MAX)

# pinned

parse_pinned = bool_parser_factory('pinned', False)
put_pinned_into = bool_optional_putter_factory('pinned', False)
validate_pinned = bool_validator_factory('pinned', False)

# reactions

def parse_reactions(data, old_reactions = None):
    """
    Parses the message's reactions.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message data.
    old_reactions : `None`, ``ReactionMapping`` = `None`, Optional
        The old reactions on the message.
    
    Returns
    -------
    reactions : `None`, ``ReactionMapping``
    """
    reactions_data = data.get('reactions', None)
    if (reactions_data is None) or (not reactions_data):
        new_reactions = None
    else:
        new_reactions = ReactionMapping.from_data(reactions_data)
    
    return merge_update_reaction_mapping(old_reactions, new_reactions)


def put_reactions_into(reactions, data, defaults):
    """
    Puts the message's reactions' data into the given `data` json serializable object.
    
    Parameters
    ----------
    reactions : `None`, ``ReactionMapping``
        The message's reactions.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (reactions is not None):
        data['reactions'] = reactions.to_data()
    elif defaults:
        data['reactions'] = []
    
    return data


def validate_reactions(reactions):
    """
    Validates message's reactions.
    
    Parameters
    ----------
    reactions : `None`, ``ReactionMapping`` (or compatible)
        Reactions value to validate.
    
    Returns
    -------
    reactions : `None`, ``ReactionMapping``
    
    Raises
    ------
    TypeError
        - `reactions` of invalid type given.
    """
    if reactions is None:
        return None
    
    if isinstance(reactions, ReactionMapping):
        return reactions
    
    return ReactionMapping(reactions)

# referenced_message

def parse_referenced_message(data):
    """
    Parses out the referenced message from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message data.
    
    Returns
    -------
    referenced_message : `None`, ``Message``
    """
    referenced_message_data = data.get('referenced_message', None)
    if (referenced_message_data is not None):
        return Message.from_data(referenced_message_data)
    
    referenced_message_data = data.get('message_reference', None)
    if (referenced_message_data is not None):
        return Message._create_from_partial_data(referenced_message_data)


def put_referenced_message_into(
    referenced_message, data, defaults, *, recursive = True, message_type = MessageType.default
):
    """
    Puts the referenced message's data into the  given `data` json serializable object.
    
    Parameters
    ----------
    referenced_message : `None`, ``Message``
        The referenced message.
    
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    recursive : `bool` = `True`, Optional (Keyword only)
        Whether the referenced message's data should be included as well.
    
    message_type : ``MessageType` = `MessageType.default`, Optional (Keyword only)
        The message's type. Used to check whether the referenced message should be recursively included.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (referenced_message is not None):
        data['message_reference'] = referenced_message.to_message_reference_data()
        
        if recursive and message_type in (MessageType.inline_reply, MessageType.thread_started):
            data['referenced_message'] = referenced_message.to_data(
                defaults = defaults, include_internals = True, recursive = False
            )
    
    return data


validate_referenced_message = nullable_entity_validator_factory(
    'referenced_message', NotImplemented, include = 'Message'
)

# role_subscription

parse_role_subscription = nullable_entity_parser_factory('role_subscription_data', MessageRoleSubscription)
put_role_subscription_into = nullable_entity_optional_putter_factory(
    'role_subscription_data', MessageRoleSubscription
)
validate_role_subscription = nullable_entity_validator_factory('role_subscription', MessageRoleSubscription)

# stickers

parse_stickers = nullable_functional_array_parser_factory(
    'sticker_items', create_partial_sticker_from_partial_data, do_sort = False
)
put_stickers_into = nullable_functional_array_optional_putter_factory('sticker_items', create_partial_sticker_data)
validate_stickers = nullable_entity_array_validator_factory(
    'stickers', Sticker, sort_key = id_sort_key,
)

# thread

def parse_thread(data, guild_id = 0):
    """
    Parses out the thread started by the message from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message data.
    guild_id : `int` = `0`, Optional
        The guild's identifier where the message was sent to.
    """
    thread_data = data.get('thread', None)
    if (thread_data is not None):
        return Channel.from_data(thread_data, guild_id = guild_id)


put_thread_into = nullable_entity_optional_putter_factory(
    'thread', NotImplemented, force_include_internals = True
)
validate_thread = nullable_entity_validator_factory('thread', NotImplemented, include = 'Channel')

# tts

parse_tts = bool_parser_factory('tts', False)
put_tts_into = bool_optional_putter_factory('tts', False)
validate_tts = bool_validator_factory('tts', False)

# type

parse_type = preinstanced_parser_factory('type', MessageType, MessageType.default)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', MessageType)



__all__ = ('RoleManagerMetadataSubscription',)

from scarletio import copy_docs

from .integration import RoleManagerMetadataIntegration
from .fields import (
    parse_purchasable, parse_subscription_listing_id, put_purchasable_into, put_subscription_listing_id_into,
    validate_purchasable, validate_subscription_listing_id
)


class RoleManagerMetadataSubscription(RoleManagerMetadataIntegration):
    """
    Role manager metadata of a role managed by a subscription integration.
    
    Attributes
    ----------
    integration_id : `int`
        The manager integration's identifier.
    purchasable : `bool`
        Whether this role is available for purchase.
    subscription_listing_id : `int`
        The subscription listing's and sku's identifier to which the role belongs to.
    """
    __slots__ = ('purchasable', 'subscription_listing_id')
    
    def __new__(cls, *, integration_id = ..., purchasable = ..., subscription_listing_id = ...):
        """
        Creates a new role manager.
        
        Parameters
        ----------
        integration_id : `int`, ``Integration``, Optional (Keyword only)
            The manager integration's identifier.
        purchasable : `bool`, Optional (Keyword only)
            Whether this role is available for purchase.
        subscription_listing_id : `int`, Optional (Keyword only)
            The subscription listing's and sku's identifier to which the role belongs to.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # purchasable
        if purchasable is ...:
            purchasable = True
        else:
            purchasable = validate_purchasable(purchasable)
        
        # subscription_listing_id
        if subscription_listing_id is ...:
            subscription_listing_id = 0
        else:
            subscription_listing_id = validate_subscription_listing_id(subscription_listing_id)
        
        self = RoleManagerMetadataIntegration.__new__(cls, integration_id = integration_id)
        self.purchasable = purchasable
        self.subscription_listing_id = subscription_listing_id
        return self
    
    
    @classmethod
    @copy_docs(RoleManagerMetadataIntegration.from_data)
    def from_data(cls, data):
        self = super(RoleManagerMetadataSubscription, cls).from_data(data)
        self.purchasable = parse_purchasable(data)
        self.subscription_listing_id = parse_subscription_listing_id(data)
        return self
    
    
    @copy_docs(RoleManagerMetadataIntegration.to_data)
    def to_data(self, *, defaults = False):
        data = RoleManagerMetadataIntegration.to_data(self)
        put_purchasable_into(self.purchasable, data, defaults)
        put_subscription_listing_id_into(self.subscription_listing_id, data, defaults)
        return data
    
    
    @copy_docs(RoleManagerMetadataIntegration.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' integration_id = ')
        repr_parts.append(repr(self.integration_id))
        
        repr_parts.append(', purchasable = ')
        repr_parts.append(repr(self.purchasable))
        
        repr_parts.append(', subscription_listing_id = ')
        repr_parts.append(repr(self.subscription_listing_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(RoleManagerMetadataIntegration.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # integration_id
        hash_value ^= self.integration_id
        
        # purchasable
        hash_value ^= self.purchasable
        
        # subscription_listing_id
        hash_value ^= self.subscription_listing_id
        
        return hash_value
    
    
    @copy_docs(RoleManagerMetadataIntegration.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # integration_id
        if self.integration_id != other.integration_id:
            return False
        
        # purchasable
        if self.purchasable != other.purchasable:
            return False
        
        # subscription_listing_id
        if self.subscription_listing_id != other.subscription_listing_id:
            return False
        
        return True
    
    
    @copy_docs(RoleManagerMetadataIntegration.copy)
    def copy(self):
        new = RoleManagerMetadataIntegration.copy(self)
        new.purchasable = self.purchasable
        new.subscription_listing_id = self.subscription_listing_id
        return new
    
    
    def copy_with(self, *, integration_id = ..., purchasable = ..., subscription_listing_id = ...):
        """
        Copies the role manager metadata with the given fields.
        
        Parameters
        ----------
        integration_id : `int`, ``Integration``, Optional (Keyword only)
            The manager integration's identifier.
        purchasable : `bool`, Optional (Keyword only)
            Whether this role is available for purchase.
        subscription_listing_id : `int`, Optional (Keyword only)
            The subscription listing's and sku's identifier to which the role belongs to.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # purchasable
        if purchasable is ...:
            purchasable = self.purchasable
        else:
            purchasable = validate_purchasable(purchasable)
        
        # subscription_listing_id
        if subscription_listing_id is ...:
            subscription_listing_id = self.subscription_listing_id
        else:
            subscription_listing_id = validate_subscription_listing_id(subscription_listing_id)
        
        new = RoleManagerMetadataIntegration.copy_with(self, integration_id = integration_id)
        new.purchasable = purchasable
        new.subscription_listing_id = subscription_listing_id
        return new



__all__ = ()

import reprlib
from json import JSONDecodeError

from scarletio import from_json
from scarletio.web_common import Formdata


TYPE_NAME_UNEXPECTED = 'object'
TYPE_NAME_STRING = 'string'
TYPE_NAME_BINARY = 'binary'
TYPE_NAME_BOOLEAN = 'boolean'
TYPE_NAME_INTEGER = 'integer'
TYPE_NAME_FLOAT = 'float'
TYPE_NAME_LIST = 'list'
TYPE_NAME_HASH_MAP = 'object'
TYPE_NAME_FORMDATA = 'formdata'

MODIFIER_LENGTH = 'length'

VALUE_INDENT = '    '
VALUE_NONE = 'null'
VALUE_BOOLEAN_TRUE = 'true'
VALUE_BOOLEAN_FALSE = 'false'

STRING_MIN_LINE_LENGTH = 60
STRING_MAX_LINE_LENGTH = 120
STRING_BREAK_TO_MULTI_LINE_OVER = 60
STRING_MAX_RENDER_LENGTH = 12000


def reconstruct_payload(payload):
    """
    Tries to reconstruct the given payload.
    
    Parameters
    ----------
    payload : `object`
        The payload to try to reconstruct.
    
    Returns
    -------
    reconstructed_value : `str`
    """
    if payload is None: # nothing to do
        return None
    
    into = ['payload = ']
    if isinstance(payload, str):
        reconstruct_json_into(payload, into, 0)
    
    elif isinstance(payload, Formdata):
        reconstruct_formdata_into(payload, into)
    
    elif isinstance(payload, bytes):
        reconstruct_binary_into(payload, into)
    
    else:
        reconstruct_unexpected_into(payload, into)
    
    return ''.join(into)


def reconstruct_json_into(value, into, indent):
    """
    Reconstructs a json payload extending the given `into` list.
    
    Parameters
    ----------
    value : `str`
        Json payload data to reconstruct.
    into : `list` of `str`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    """
    try:
        json_data = from_json(value)
    except JSONDecodeError:
        into.append(TYPE_NAME_STRING)
        into.append('(')
        into.append(MODIFIER_LENGTH)
        into.append('=')
        into.append(str(len(value)))
        into.append('): ')
        into.append(reprlib.repr(value))
    else:
        reconstruct_value_into(json_data, into, indent, False)


def reconstruct_value_into(value, into, indent, is_file):
    """
    Reconstructs a value extending the given `into` list.
    
    Parameters
    ----------
    value : `None`, `str`, `list`, `dict`, `int`, `float`, `bool`
        The deserialized json value.
    into : `list` of `str`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    is_file : `bool`
        Whether the value is a file and should not shown.
    """
    if value is None:
        reconstruct_null_into(into)
        return
    
    if isinstance(value, bool):
        reconstruct_boolean_into(value, into)
        return

    if isinstance(value, str):
        reconstruct_string_into(value, into, indent, is_file)
        return
    
    if isinstance(value, int):
        reconstruct_integer_into(value, into)
        return
    
    if isinstance(value, float):
        reconstruct_float_into(value, into)
        return
    
    if isinstance(value, list):
        reconstruct_list_into(value, into, indent)
        return
    
    if isinstance(value, dict):
        reconstruct_hash_map_into(value, into, indent)
        return
    
    # Used at form data
    if isinstance(value, bytes):
        reconstruct_binary_into(value, into)
        return
    
    reconstruct_unexpected_into(value, into)


def reconstruct_null_into(into):
    """
    Reconstructs a null value to the given `into` list.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to extend it's content.
    """
    into.append(VALUE_NONE)


def reconstruct_boolean_into(value, into):
    """
    Reconstructs a boolean value to the given `into` list.
    
    Parameters
    ----------
    value : `bool`
        The boolean value.
    into : `list` of `str`
        A list to extend it's content.
    """
    if value:
        value_representation = VALUE_BOOLEAN_TRUE
    else:
        value_representation = VALUE_BOOLEAN_FALSE
    
    into.append(value_representation)


def reconstruct_string_into(value, into, indent, is_file):
    """
    Reconstructs a string value to the given `into` list.
    
    Parameters
    ----------
    value : `str`
        The string value.
    into : `list` of `str`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    is_file : `bool`
        Whether the value is a file and should not be shown.
    """
    length = len(value)
    
    if is_file or (length > STRING_MAX_RENDER_LENGTH):
        into.append(TYPE_NAME_STRING)
        into.append('(')
        into.append(MODIFIER_LENGTH)
        into.append('=')
        into.append(str(length))
        into.append(')')
    
    else:
        if length > STRING_BREAK_TO_MULTI_LINE_OVER:
            into.append('(\n')
            
            string_indent = indent + 1
            
            chunk_size = STRING_MAX_LINE_LENGTH - string_indent * len(VALUE_INDENT)
            if chunk_size < STRING_MIN_LINE_LENGTH:
                chunk_size = STRING_MIN_LINE_LENGTH
            
            start_index = 0
            while True:
                end_index = start_index + chunk_size
                if end_index >= length:
                    end_index = length
                    should_break = True
                else:
                    should_break = False
                
                for counter in range(string_indent):
                    into.append(VALUE_INDENT)
                
                into.append(repr(value[start_index:end_index]))
                into.append('\n')
                
                if should_break:
                    break
                
                start_index = end_index
                continue
            
            for counter in range(indent):
                into.append(VALUE_INDENT)
            
            into.append(')')
        
        else:
            into.append(repr(value))


def reconstruct_integer_into(value, into):
    """
    Reconstructs an integer value to the given `into` list.
    
    Parameters
    ----------
    value : `int`
        The integer value.
    into : `list` of `str`
        A list to extend it's content.
    """
    into.append(repr(value))


def reconstruct_float_into(value, into):
    """
    Reconstructs a float value to the given `into` list.
    
    Parameters
    ----------
    value : `float`
        The float value.
    into : `list` of `str`
        A list to extend it's content.
    """
    into.append(repr(value))


def reconstruct_list_into(value, into, indent):
    """
    Reconstructs a list value extending the given `into` list.
    
    Parameters
    ----------
    value : `list`
        The list value.
    into : `list` of `str`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    """
    length = len(value)
    
    into.append('[')
    
    if length:
        into.append('\n')
        
        element_indent = indent + 1
        
        for index, list_element in enumerate(value):
            for counter in range(element_indent):
                into.append(VALUE_INDENT)
            into.append(str(index))
            into.append(': ')
            
            reconstruct_value_into(list_element, into, element_indent, False)
            
            into.append(',\n')
        
        for counter in range(indent):
            into.append(VALUE_INDENT)
    
    into.append(']')


def reconstruct_hash_map_into(value, into, indent):
    """
    Reconstructs a hash map value extending the given `into` list.
    
    Parameters
    ----------
    value : `dict`
        The hash map value.
    into : `list` of `str`
        A list to extend it's content.
    indent : `int`
        The amount of indents to add.
    """
    length = len(value)
    
    into.append('{')
    
    if length:
        into.append('\n')
        item_indent = indent + 1
        
        for item_key, item_value in sorted(value.items(), key = hash_map_key_sort_key):
            for counter in range(item_indent):
                into.append(VALUE_INDENT)
            
            into.append(repr(item_key))
            into.append(': ')
            
            reconstruct_value_into(item_value, into, item_indent, False)
            
            into.append(',\n')
        
        for counter in range(indent):
            into.append(VALUE_INDENT)
    
    into.append('}')


def reconstruct_unexpected_into(value, into):
    """
    Reconstructs a value with an unexpected type to the given `into` list.
    
    Parameters
    ----------
    value : `object`
        Le value.
    into : `list` of `str`
        A list to extend it's content.
    """
    into.append(TYPE_NAME_UNEXPECTED)
    into.append(': ')
    into.append(reprlib.repr(value))


def reconstruct_binary_into(value, into):
    """
    Reconstructs a binary value to the given `into` list.
    
    Parameters
    ----------
    value : `binary`
        The binary value.
    into : `list` of `str`
        A list to extend it's content.
    """
    into.append(TYPE_NAME_BINARY)
    into.append('(')
    into.append(MODIFIER_LENGTH)
    into.append('=')
    into.append(str(len(value)))
    into.append(')')


def reconstruct_formdata_into(value, into):
    """
    Tries to reconstruct formdata into the given `into` list.
    
    Parameters
    ----------
    value : ``Formdata``
        The formdata to reconstruct.
    into : `list` of `str`
        A list to extend it's content.
    
    Returns
    -------
    reconstructed_value : `str`
    """
    into.append(TYPE_NAME_FORMDATA)
    
    fields = value.fields
    
    into.append('({\n')
    
    for index, (field_type_options, field_headers, field_value) in enumerate(fields):
        field_name = field_type_options['name']
        filename = field_type_options.get('filename', None)
        
        into.append(VALUE_INDENT)
        into.append(str(index))
        into.append(': ')
        into.append(repr(field_name))
        if (filename is not None) and (filename != field_name):
            into.append(' | ')
            into.append(repr(filename))
        
        into.append(': ')
        if (filename is None) and (field_name == 'payload_json'):
            reconstruct_json_into(field_value, into, 1)
        else:
            reconstruct_value_into(field_value, into, 1, True)
        into.append('\n')
    
    into.append('})')


def hash_map_key_sort_key(item):
    """
    Used to sort hash maps based on their key.
    
    Parameters
    ----------
    item : `tuple` (`str`, `object`)
        An item of a hash map.
    
    Returns
    -------
    key : `str`
    """
    return item[0]



__all__ = ('ApplicationCommandOptionMetadataParameter',)

from scarletio import copy_docs

from .base import ApplicationCommandOptionMetadataBase
from .fields import parse_required, put_required_into, validate_required


class ApplicationCommandOptionMetadataParameter(ApplicationCommandOptionMetadataBase):
    """
    Base type for parameter application command option metadatas.
    
    Attributes
    ----------
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    """
    __slots__ = ('required',)
    
    
    def __new__(cls, *, required = ...):
        """
        Creates a new parameter application command option metadata with the given parameters.
        
        Parameters
        ----------
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # required
        if required is ...:
            required = False
        else:
            required = validate_required(required)
        
        # Construct
        new = object.__new__(cls)
        new.required = required
        return new
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            required = keyword_parameters.pop('required', ...),
        )
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.required = parse_required(data)
        return self
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_required_into(self.required, data, defaults)
        return data
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase._add_type_specific_repr_fields)
    def _add_type_specific_repr_fields(self, repr_parts):
        repr_parts.append(' required = ')
        repr_parts.append(repr(self.required))
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if self.required != other.required:
            return False
        
        return True
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # required
        hash_value ^= self.required << 1
        
        return hash_value
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.required = self.required
        return new
    
    
    def copy_with(self, *, required = ...):
        """
        Copies the parameter application command option metadata with the given fields.
        
        Parameters
        ----------
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # required
        if required is ...:
            required = self.required
        else:
            required = validate_required(required)
        
        # Construct
        new = object.__new__(type(self))
        new.required = required
        return new
    
    
    @copy_docs(ApplicationCommandOptionMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            required = keyword_parameters.pop('required', ...),
        )



__all__ = ('FlagBase', 'ReverseFlagBase', )

import warnings


def maybe_apply_deprecation(function, name, deprecation_info):
    """
    CApplies deprecation to the given getter / enabled / disabler function.
    
    Parameters
    ----------
    function : `FunctionType`
        The function to apply it to.
    name : `str`
        The flag's name.
    deprecation_info : `None`, `tuple` (`str`, `str`)
        Deprecation info for the field if deprecated.
    
    Returns
    -------
    function : `FunctionType`
    """
    if (deprecation_info is None):
        return function
    
    def deprecated_field(self):
        nonlocal deprecation_info
        nonlocal function
        nonlocal name
        
        warnings.warn(
            (
                f'{self.__class__.__name__}`\s {name} field is deprecated and will be removed in '
                f'{deprecation_info[0]}. Please use {deprecation_info[1]} instead.'
            ),
            FutureWarning,
            stacklevel = 3,
        )
        
        return function(self)
    
    return deprecated_field


def create_flag_getter(name, shift, deprecation_info):
    """
    Creates a flag getter function.
    
    Parameters
    ----------
    name : `str`
        The flag's name.
    shift : `int`
        Bit shift value.
    deprecation_info : `None`, `tuple` (`str`, `str`)
        Deprecation info for the field if deprecated.
    
    Returns
    -------
    flag_getter : `FunctionType`
    """
    def flag_getter(self):
        nonlocal shift
        return (self >> shift) & 1
    
    maybe_apply_deprecation(flag_getter, name, deprecation_info)
    function_name = f'_get_{name}'
    flag_getter.__name__ = function_name
    flag_getter.__qualname__ = function_name
    
    return flag_getter


def create_reversed_flag_getter(name, shift, deprecation_info):
    """
    Creates a reversed flag getter function.
    
    Parameters
    ----------
    name : `str`
        The flag's name.
    shift : `int`
        Bit shift value.
    deprecation_info : `None`, `tuple` (`str`, `str`)
        Deprecation info for the field if deprecated.
    
    Returns
    -------
    reversed_flag_getter : `FunctionType`
    """
    def flag_getter(self):
        nonlocal shift
        return ((self >> shift) & 1) ^ 1
    
    maybe_apply_deprecation(flag_getter, name, deprecation_info)
    function_name = f'_get_{name}'
    flag_getter.__name__ = function_name
    flag_getter.__qualname__ = function_name
    
    return flag_getter



def create_flag_enabler(name, shift, deprecation_info):
    """
    Creates a flag enabler function.
    
    Parameters
    ----------
    name : `str`
        The flag's name.
    shift : `int`
        Bit shift value.
    deprecation_info : `None`, `tuple` (`str`, `str`)
        Deprecation info for the field if deprecated.
    
    Returns
    -------
    flag_enabler : `FunctionType`
    """
    def flag_enabled(self):
        nonlocal shift
        return int.__new__(type(self), (self | (1 << shift)))
    
    maybe_apply_deprecation(flag_enabled, name, deprecation_info)
    function_name = f'_enable_{name}'
    flag_enabled.__name__ = function_name
    flag_enabled.__qualname__ = function_name
    
    return flag_enabled


def create_flag_disabler(name, shift, deprecation_info):
    """
    Creates a flag disabler function.
    
    Parameters
    ----------
    name : `str`
        The flag's name.
    shift : `int`
        Bit shift value.
    deprecation_info : `None`, `tuple` (`str`, `str`)
        Deprecation info for the field if deprecated.
    
    Returns
    -------
    flag_disabler : `FunctionType`
    """
    def flag_disabler(self):
        nonlocal shift
        if (self >> shift) & 1:
            return int.__new__(type(self), (self ^ (1 << shift)))
        
        return self
    
    maybe_apply_deprecation(flag_disabler, name, deprecation_info)
    function_name = f'_disable_{name}'
    flag_disabler.__name__ = function_name
    flag_disabler.__qualname__ = function_name
    
    return flag_disabler


def validate_shift(shift, keys_name, shift_name):
    """
    Validates a shift.
    
    shift : `int`
        The shift to validate.
    keys_name : `str`
        The name of the keys we are validating.
    shift_name : `str`
        The shift field's name.
    
    Raises
    ------
        - If `shift`'s type or value is not acceptable.
    """
    if not isinstance(shift, int):
        raise TypeError(
            f'`{keys_name}`\'s {shift_name}s should be `int`, got {shift.__class__.__name__}; {shift!r}.'
        )
    
    if shift < 0 or shift > 63:
        raise TypeError(
            f'`{keys_name}`\' {shift_name}s must be between 0 and 63, got: {shift!r}'
        )


def deprecated_value_validator(value, keys_name):
    """
    Deprecated value validator for key values.
    
    Parameters
    ----------
    value : `int`
        The value to validate.
    keys_name : `str`
        The name of the keys we are validating.
    
    Raises
    ------
        - If `value`'s type or structure is not acceptable.
    """
    if not isinstance(value, tuple) or (len(value) != 3):
        raise TypeError(
            f'`{keys_name}`\'s values can be `tuple` of length 3.'
        )
    
    shift, removed_at, use_instead = value
    validate_shift(shift, keys_name, 'shift')
    
    if not isinstance(removed_at, str):
        raise TypeError(
            f'`{keys_name}`\'s removed_at-s should be `str`, got {removed_at.__class__.__name__}; {removed_at!r}.'
        )
    
    if not isinstance(use_instead, str):
        raise TypeError(
            f'`{keys_name}`\'s use_instead-s should be `str`, got {removed_at.__class__.__name__}; {removed_at!r}.'
        )


def default_value_validator(value, keys_name):
    """
    Default validator for key values.
    
    Parameters
    ----------
    value : `int`
        The value to validate.
    keys_name : `str`
        The name of the keys we are validating.
    
    Raises
    ------
        - If `value`'s type or value is not acceptable.
    """
    validate_shift(value, keys_name, 'value')


def validate_keys(keys, keys_name, value_validator):
    """
    Validates the given flag keys.
    
    Parameters
    ----------
    keys : `dict` of (`str`, `int`) items
        Flag keys to validate.
    keys_name : `str`
        The name of the keys we are validating.
    value_validator : `FunctionType`
        Validates the item's value.
    
    Raises
    ------
        - If `keys`' type or structure is not acceptable.
    """
    if not isinstance(keys, dict):
        raise TypeError(
            f'`{keys_name}` defined as non `dict`: `{keys.__class__.__name__}`.'
        )
    
    for name, value in keys.items():
        if not isinstance(name, str):
            raise TypeError(
                f'`{keys_name}`\'s keys should be `str`-s, meanwhile got at least 1 non `str`: '
                f'{name.__class__.__name__}; {name!r}.'
            )
        
        value_validator(value, keys_name)


def iterate_keys(keys, deprecated_keys):
    """
    Iterates over the items of the given keys.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    keys : `dict` of (`str`, `int`) items
        Flag keys to iterate over.
    validate_keys : `NotImplemented`, `dict` of (`str`, `int`) items
        Deprecated flag keys to validate.
    
    Yields
    ------
    key : `str`
        Field key.
    shift : `int`
        Flag shift.
    deprecation_info : `None`, `tuple` (`str`, `str`)
        The deprecation info of the field.
    """
    for item in keys.items():
        yield (*item, None)
    
    if (deprecated_keys is not NotImplemented):
        for key, (shift, *deprecation_info) in deprecated_keys.items():
            yield key, shift, deprecation_info


class FlagMeta(type):
    """
    Metaclass for bitwise flags.
    """
    def __new__(
        cls,
        class_name,
        class_parents,
        class_attributes,
        access_keyword = None,
        enable_keyword = None,
        disable_keyword = None,
        base_class = False,
    ):
        """
        Creates a bitwise flag type.
        
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type`
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        access_keyword : `str`, Optional
            The string what will go before the descriptor's name, which allow accessing their respective value. If not
            defined, the there will be no keyword before them.
        enable_keyword : `str`, Optional
            Whether enabling descriptors should be added to the created type. The passed string will be before the
            enabling descriptor's name. If not defined, then the created type will not have flag enablers.
        disable_keyword : `str`, Optional
            Whether disabling descriptors should be added to the created type. The passed string will be before the
            disabling descriptor's name. If not defined, then the created type will not have flag disablers.
        base_class : `bool`, Optional
            Whether the created type is a base class for flags. If it is:
                - Should directly derived from `int`.
                - Should not implement `__keys__` (except `NotImplemented`).
                - Should implement `__getter_factory__` class attribute.
                - Should implement `__enabler_factory__` class attribute.
                - Should implement `__disabler_factory__` class attribute.
            If not, then:
                - Should derive directly from a ``FlagMeta`` base_class.
                - Should implement `__keys__` as `dict` of (`str`, `int`) items, where the values are not less than
                    `0`, and not greater than `63` either.
        
        Returns
        -------
        type : ``FlagMeta``
        
        Raises
        ------
        TypeError
            When any requirements are not satisfied.
        """
        if base_class:
            if (not class_parents) or (not issubclass(class_parents[0], int)):
                raise TypeError(
                    f'`{class_name}` is not derived directly from `int`.'
                )
            
            class_keys = class_attributes.get('__keys__', ...)
            if class_keys is NotImplemented:
                pass
            elif class_keys is ...:
                class_attributes['__keys__'] = NotImplemented
            else:
                raise TypeError(
                    f'`{class_name}` has `__keys__` defined and not as `NotImplemented`.'
                )
            
            for attribute_name in ('__getter_factory__', '__enabler_factory__', '__disabler_factory__'):
                if attribute_name not in class_attributes:
                    raise TypeError(
                        f'`{class_name}` should implement a `{attribute_name}`.'
                    )
            
            # do not care about the leftover
            
            return type.__new__(cls, class_name, class_parents, class_attributes)
        
        # Python has no GOTO, so lets insert one
        while True:
            if class_parents:
                parent = class_parents[0]
                if (type(parent) is FlagMeta) and parent.__keys__ is NotImplemented:
                    break
            
            raise TypeError(
                f'`{class_name}` is not derived directly from a `{cls.__name__}` base instance.'
            )
        
        # Validate keys
        try:
            keys = class_attributes['__keys__']
        except KeyError:
            raise TypeError(
                f'`{class_name}` did not define `__keys__` attribute.'
            ) from None
        
        validate_keys(keys, '__keys__', default_value_validator)
        
        deprecated_keys = class_attributes.get('__deprecated_keys__', NotImplemented)
        if (deprecated_keys is not NotImplemented):
            validate_keys(deprecated_keys, '__deprecated_keys__', deprecated_value_validator)
        
        class_attributes.setdefault('__new__', int.__new__)
        
        getter = parent.__getter_factory__
        enabler = parent.__enabler_factory__
        disabler = parent.__disabler_factory__
        
        # Add properties
        for name, shift, deprecation_info in iterate_keys(keys, deprecated_keys):
            if access_keyword is None:
                access_name = name
            else:
                access_name = f'{access_keyword}_{name}'
            
            class_attributes[access_name] = property(getter(name, shift, deprecation_info))
            
            if (enable_keyword is not None):
                class_attributes[f'{enable_keyword}_{name}'] = enabler(name, shift, deprecation_info)
            
            if (disable_keyword is not None):
                class_attributes[f'{disable_keyword}_{name}'] = disabler(name, shift, deprecation_info)
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


class FlagBase(int, metaclass = FlagMeta, base_class = True):
    """
    Base class for bitwise flags.
    
    Class Attributes
    ----------------
    __getter_factory__ : ``create_flag_getter``
        Flag value getter descriptor for subclasses.
    __enabler_factory__ : ``create_flag_enabler``
        Flag enabler function factory for subclasses.
    __disabler_factory__ : ``create_flag_disabler``
        Flag disabler function factory for subclasses.
    """
    __slots__ = ()
    __keys__ = NotImplemented
    __deprecated_keys__ = NotImplemented
    
    __getter_factory__ = create_flag_getter
    __enabler_factory__ = create_flag_enabler
    __disabler_factory__ = create_flag_disabler
    
    def __new__(self, base = None):
        """You cannot subclass flag base classes."""
        raise NotImplementedError
    
    def __repr__(self):
        """Returns the representation of the flag."""
        return f'{self.__class__.__name__}({int.__repr__(self)})'
    
    def __getitem__(self, key):
        """Returns whether a specific flag of the given name is enabled."""
        return (self >> self.__keys__[key]) & 1
    
    def keys(self):
        """
        Yields the name of the bitwise flags, which are enabled.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
        """
        for name, shift in self.__keys__.items():
            if (self >> shift) & 1:
                yield name
    
    __iter__ = keys
    
    def values(self):
        """
        Yields the shift values of the flags, under which shift value the flag is enabled.
        
        This method is a generator.
        
        Yields
        -------
        shift : `int`
        """
        for shift in self.__keys__.values():
            if (self >> shift) & 1:
                yield shift
    
    def items(self):
        """
        Yields the items of the flag.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
            The name of the specific flag
        enabled : `int` (`0`, `1`)
            Whether the specific bitwise value is enabled.
        """
        for name, shift in self.__keys__.items():
            yield name, (self >> shift) & 1
    
    def __contains__(self, key):
        """Returns whether the specific flag of the given name is enabled."""
        try:
            position = self.__keys__[key]
        except KeyError:
            return 0
        
        return (self >> position) & 1
    
    def is_subset(self, other):
        """Returns whether self has the same amount or more flags disabled than other."""
        return (self & other) == self
    
    def is_superset(self, other):
        """Returns whether self has the same amount or more flags enabled than other."""
        return (self | other) == self
    
    def is_strict_subset(self, other):
        """Returns whether self has more flags disabled than other."""
        return self != other and (self & other) == self
    
    def is_strict_superset(self, other):
        """Returns whether self has more flags enabled than other."""
        return self != other and (self | other) == self
    
    __ge__ = is_superset
    __gt__ = is_strict_superset
    __lt__ = is_strict_subset
    __le__ = is_subset
    
    
    def _get_shift_of(self, key):
        """
        Gets the shift value for the given keys.
        
        Parameters
        ----------
        keys : `str`
            The key's name.
        
        Returns
        -------
        shift : `int`
        
        Raises
        ------
        LookupError
            - Invalid key given.
        """
        try:
            shift = self.__keys__[key]
        except KeyError:
            pass
        
        else:
            return shift
        
        deprecated_keys = self.__deprecated_keys__
        if (deprecated_keys is not NotImplemented):
            try:
                shift, removed_at, use_instead = deprecated_keys[key]
            except KeyError:
                pass
            else:
                warnings.warn(
                    (
                        f'`{self.__class__.__name__}`\'s {key} is deprecated and will be removed at {removed_at}. '
                        f'Please use {use_instead} instead.'
                    ),
                    FutureWarning,
                    stacklevel = 3,
                )
                
                return shift
        
        # support `_(\d+)` format
        if key.startswith('_'):
            try:
                shift = int(key[1:])
            except ValueError:
                pass
            else:
                return shift
        
        raise LookupError(f'Invalid key: {key!r}.')
    
    
    def update_by_keys(self, **keyword_parameters):
        """
        Updates the source value with the given flags and returns a new one.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            `flag-name` - `bool` relations.
        
        Returns
        -------
        flag : ``FlagBase``
        
        Raises
        ------
        LookupError
            If a keyword is invalid.
        
        Examples
        -------
        ```py
        >>> from hata import Permission
        >>> perm = Permission().update_by_keys(kick_users = True, ban_users = True)
        >>> list(perm)
        ['kick_users', 'ban_users']
        >>> perm = perm.update_by_keys(manage_roles = True, kick_users = False)
        >>> list(perm)
        ['ban_users', 'manage_roles']
        ```
        """
        new = self
        for key, value in keyword_parameters.items():
            shift = self._get_shift_of(key)
            
            if value:
                new |= (1 << shift)
            else:
                if (new >> shift) & 1:
                    new ^= (1 << shift)
        
        return int.__new__(type(self), new)


class ReverseFlagBase(FlagBase, base_class = True):
    """
    Base class for reversed bitwise flags.
    
    Class Attributes
    ----------------
    __getter_factory__ : ``create_reversed_flag_getter``
        Flag value getter descriptor for subclasses.
    __enabler_factory__ : ``create_flag_disabler``
        Flag enabler function factory for subclasses.
    __disabler_factory__ : ``create_flag_enabler``
        Flag disabler function factory for subclasses.
    """
    __getter_factory__ = create_reversed_flag_getter
    __enabler_factory__ = create_flag_disabler
    __disabler_factory__ = create_flag_enabler
    
    def __getitem__(self, key):
        """Returns whether a specific flag of the given name is enabled."""
        return ((self >> self.__keys__[key]) & 1)^1
    
    def keys(self):
        """
        Yields the name of the bitwise flags, which are enabled.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
        """
        for name, shift in self.__keys__.items():
            if ((self >> shift) & 1) ^ 1:
                yield name
    
    __iter__ = keys
    
    def values(self):
        """
        Yields the shift values of the flags, under which shift value the flag is enabled.
        
        This method is a generator.
        
        Yields
        -------
        shift : `int`
        """
        for shift in self.__keys__.values():
            if ((self >> shift) & 1)^1:
                yield shift
    
    def items(self):
        """
        Yields the items of the flag.
        
        This method is a generator.
        
        Yields
        -------
        name : `str`
            The name of the specific flag
        enabled : `int` (`0`, `1`)
            Whether the specific bitwise value is enabled.
        """
        for name, shift in self.__keys__.items():
            yield name, ((self >> shift) & 1) ^ 1
    
    
    def __contains__(self, key):
        """Returns whether the specific flag of the given name is enabled."""
        try:
            position = self.__keys__[key]
        except KeyError:
            return 0
        
        return ((self >> position) & 1) ^ 1
    
    
    def is_subset(self, other):
        """Returns whether self has the same amount or more flags disabled than other."""
        return (self | other) == self
    
    
    def is_superset(self, other):
        """Returns whether self has the same amount or more flags enabled than other."""
        return (self & other) == self
    
    
    def is_strict_subset(self, other):
        """Returns whether self has more flags disabled than other."""
        return self != other and (self | other) == self
    
    
    def is_strict_superset(self, other):
        """Returns whether self has more flags enabled than other."""
        return self != other and (self & other) == self
    
    __ge__ = is_superset
    __gt__ = is_strict_superset
    __lt__ = is_strict_subset
    __le__ = is_subset
    
    
    def update_by_keys(self, **keyword_parameters):
        """
        Updates the source value with the given flags and returns a new one.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            `flag-name` - `bool` relations.
        
        Returns
        -------
        flag : ``ReverseFlagBase``
        
        Examples
        --------
        ```py
        >>> from hata import SystemChannelFlag
        >>> flags = SystemChannelFlag()
        >>> list(flags)
        ['welcome', 'boost']
        >>> flags = flags.update_by_keys(boost=False)
        >>> list(flags)
        ['welcome']
        ```
        """
        new = self
        for key, value in keyword_parameters.items():
            shift = self._get_shift_of(key)
            
            if value:
                if (new >> shift) & 1:
                    new ^= (1 << shift)
            else:
                new |= (1 << shift)
        
        return int.__new__(type(self), new)



__all__ = ('AutoModerationRuleTriggerMetadataKeywordPreset', )

from scarletio import copy_docs

from .base import AutoModerationRuleTriggerMetadataBase
from .fields import (
    parse_excluded_keywords, parse_keyword_presets, put_excluded_keywords_into, put_keyword_presets_into,
    validate_excluded_keywords, validate_keyword_presets
)


class AutoModerationRuleTriggerMetadataKeywordPreset(AutoModerationRuleTriggerMetadataBase):
    """
    Keyword keyword preset trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    excluded_keywords : `None`, `tuple` of `str`
        Excluded keywords from under the rule.
    
    keyword_presets : `None`, `tuple` of ``AutoModerationKeywordPresetType``
        Keyword preset defined by Discord which will be searched for in content.
    """
    __slots__ = ('excluded_keywords', 'keyword_presets',)
    
    def __new__(cls, keyword_presets = ..., excluded_keywords = ...):
        """
        Creates a new keyword preset trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        keyword_presets : `None`, `int`, ``AutoModerationKeywordPresetType``, \
                `iterable` of (`int`, ``AutoModerationKeywordPresetType``), Optional
            Keyword preset defined by Discord which will be searched for in content.
        
        excluded_keywords : `None`, `str`, `iterable` of `str`, Optional
            Excluded keywords from the preset.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        # excluded_keywords
        if excluded_keywords is ...:
            excluded_keywords = None
        else:
            excluded_keywords = validate_excluded_keywords(excluded_keywords)
        
        # keyword_presets
        if keyword_presets is ...:
            keyword_presets = None
        else:
            keyword_presets = validate_keyword_presets(keyword_presets)
        
        # Construct
        self = object.__new__(cls)
        self.excluded_keywords = excluded_keywords
        self.keyword_presets = keyword_presets
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # keyword_presets
        keyword_presets = self.keyword_presets
        repr_parts.append(' keywords = [')
        if (keyword_presets is not None):
            index = 0
            limit = len(keyword_presets)
            while True:
                keyword = keyword_presets[index]
                repr_parts.append(repr(keyword.name))
                repr_parts.append(' ~ ')
                repr_parts.append(repr(keyword.value))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        repr_parts.append(', excluded_keywords = [')
        if (excluded_keywords is not None):
            index = 0
            limit = len(excluded_keywords)
            while True:
                excluded_keyword = excluded_keywords[index]
                repr_parts.append(repr(excluded_keyword))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.excluded_keywords = parse_excluded_keywords(data)
        self.keyword_presets = parse_keyword_presets(data)
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_excluded_keywords_into(self.excluded_keywords, data, defaults)
        put_keyword_presets_into(self.keyword_presets, data, defaults)
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # excluded_keywords
        if self.excluded_keywords != other.excluded_keywords:
            return False
        
        # keyword_presets
        if self.keyword_presets != other.keyword_presets:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            hash_value ^= len(excluded_keywords)
            
            for excluded_keyword in excluded_keywords:
                hash_value ^= hash(excluded_keyword)
        
        # keyword_presets
        keyword_presets = self.keyword_presets
        if (keyword_presets is not None):
            hash_value ^= len(keyword_presets)
            
            shift = 0
            for keyword_preset in keyword_presets:
                shift += 4
                hash_value ^= keyword_preset.value << shift
        
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            excluded_keywords = (*excluded_keywords,)
        new.excluded_keywords = excluded_keywords
        
        # keyword_presets
        keyword_presets = self.keyword_presets
        if (keyword_presets is not None):
            keyword_presets = (*keyword_presets,)
        new.keyword_presets = keyword_presets
        
        return new


    
    def copy_with(self, *, excluded_keywords = ..., keyword_presets = ...):
        """
        Copies the trigger metadata with altering it's attributes based on the given fields.
        
        Parameters
        ----------
        excluded_keywords : `None`, `str`, `iterable` of `str, Optional (Keyword only)
            Excluded keywords from the rule.
        
        keyword_presets : `None`, `int`, ``AutoModerationKeywordPresetType``, \
                `iterable` of (`int`, ``AutoModerationKeywordPresetType``), Optional (Keyword only)
            Keyword preset defined by Discord which will be searched for in content.
        
        Returns
        -------
        new : `instance<type<self>>`
               
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        """
        # excluded_keywords
        if excluded_keywords is ...:
            excluded_keywords = self.excluded_keywords
            if (excluded_keywords is not None):
                excluded_keywords = (*excluded_keywords, )
        else:
            excluded_keywords = validate_excluded_keywords(excluded_keywords)
        
        # keyword_presets
        if keyword_presets is ...:
            keyword_presets = self.keyword_presets
            if (keyword_presets is not None):
                keyword_presets = (*keyword_presets, )
        else:
            keyword_presets = validate_keyword_presets(keyword_presets)
        
        new = object.__new__(type(self))
        new.excluded_keywords = excluded_keywords
        new.keyword_presets = keyword_presets
        return new



__all__ = ('ChannelMetadataGuildStage',)

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import (
    PERMISSION_MASK_CONNECT, PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_TEXT_DENY,
    PERMISSION_VOICE_DENY_CONNECTION
)

from .fields import parse_topic, put_topic_into, validate_topic

from .guild_voice_base import ChannelMetadataGuildVoiceBase


class ChannelMetadataGuildStage(ChannelMetadataGuildVoiceBase):
    """
    Guild voice channel metadata.
    
    Attributes
    ----------
    _cache_permission : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    bitrate : `int`
        The bitrate (in bits) of the voice channel.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites :`None`,  `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    region : ``VoiceRegion``
        The voice region of the channel.
    topic : `None`, `str`
        The channel's topic.
    user_limit : `int`
        The maximal amount of users, who can join the voice channel, or `0` if unlimited.
    
    Class Attributes
    ----------------
    order_group: `int` = `2`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('topic',)
    
    
    def __new__(
        cls,
        *,
        bitrate = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        user_limit = ...,
        topic = ...,
    ):
        """
        Creates a new guild voice channel metadata from the given parameters.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        region : ``VoiceRegion``, `str`, Optional (Keyword only)
            The voice region of the channel.
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
            
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # topic
        if topic is ...:
            topic = None
        else:
            topic = validate_topic(topic)
        
        # Construct
        self = ChannelMetadataGuildVoiceBase.__new__(
            cls,
            bitrate = bitrate,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
            region = region,
            user_limit = user_limit,
        )
        self.topic = topic
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
            topic = keyword_parameters.pop('topic', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildVoiceBase.__hash__(self)
        
        # topic
        topic = self.topic
        if (topic is not None):
            hash_value ^= hash(topic)
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildVoiceBase._is_equal_same_type(self, other):
            return False
        
        # topic
        if self.topic != other.topic:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._get_display_name)
    def _get_display_name(self):
        return self.name.upper()
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildStage, cls)._create_empty()
        
        self.topic = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.copy)
    def copy(self):
        new = ChannelMetadataGuildVoiceBase.copy(self)
        new.topic = self.topic
        return new
    
    
    def copy_with(
        self,
        *,
        bitrate = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        user_limit = ...,
        topic = ...,
    ):
        """
        Copies the guild voice channel metadata with the given fields.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        region : ``VoiceRegion``, `str`, Optional (Keyword only)
            The voice region of the channel.
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # topic
        if topic is ...:
            topic = self.topic
        else:
            topic = validate_topic(topic)
        
        # Construct
        new = ChannelMetadataGuildVoiceBase.copy_with(
            self,
            bitrate = bitrate,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
            region = region,
            user_limit = user_limit,
        )
        new.topic = topic
        return new
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
            topic = keyword_parameters.pop('topic', ...),
        )
    
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildVoiceBase._update_attributes(self, data)
        
        # topic
        self.topic = parse_topic(data)
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildVoiceBase._difference_update_attributes(self, data)
        
        # topic
        topic = parse_topic(data)
        if self.topic != topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildVoiceBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # topic
        put_topic_into(self.topic, data, defaults)
        
        return data
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have text permissions
        result &= PERMISSION_TEXT_DENY
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have text permissions
        result &= PERMISSION_TEXT_DENY
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)



__all__ = (
    'parse_all_emojis', 'parse_all_emojis_ordered', 'parse_custom_emojis', 'parse_custom_emojis_ordered', 'parse_emoji',
    'parse_reaction'
)

from ...core import BUILTIN_EMOJIS, UNICODE_TO_EMOJI
from ...utils import EMOJI_RP, REACTION_RP

from ..emoji import Emoji
from ..unicode.unicode_type import VARIATION_SELECTOR_16_POSTFIX

from .pattern import EMOJI_ALL_RP


VARIATION_SELECTOR_16_POSTFIX_WITH_COLON = VARIATION_SELECTOR_16_POSTFIX + ':'


def parse_emoji(text):
    """
    Tries to parse out an ``Emoji`` from the inputted text. This emoji can be custom and unicode emoji as well.
    
    If the parsing fails the function returns `None`.
    
    Parameters
    ----------
    text : `str`
        The text to parse emojis from.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    parsed = EMOJI_RP.fullmatch(text)
    if (parsed is not None):
        animated, name, emoji_id = parsed.groups()
        animated = (animated is not None)
        emoji_id = int(emoji_id)
        return Emoji._create_partial(emoji_id, name, animated)
        
    try:
        return UNICODE_TO_EMOJI[text]
    except KeyError:
        pass
    
    if text.startswith(':') and text.endswith(':') and not text.endswith(VARIATION_SELECTOR_16_POSTFIX_WITH_COLON):
        try:
            return BUILTIN_EMOJIS[text[1:-1]]
        except KeyError:
            pass
    
    return None


def _iter_parse_custom_emojis(text):
    """
    Iterates over all the custom emojis in the text as they appear.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Yields
    ------
    emoji : ``Emoji``
    """
    for groups in EMOJI_RP.findall(text):
        
        animated, name, emoji_id = groups
        animated = (True if animated else False)
        emoji_id = int(emoji_id)
        
        yield Emoji._create_partial(emoji_id, name, animated)


def _iter_parse_all_emojis(text):
    """
    Iterates over all emojis in the text as they appear.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Yields
    ------
    emoji : ``Emoji``
    """
    for groups in EMOJI_ALL_RP.findall(text):
        
        unicode_value, unicode_name, custom_animated, custom_name, custom_emoji_id = groups
        if unicode_value:
            yield UNICODE_TO_EMOJI[unicode_value]
            continue
        
        if unicode_name:
            yield BUILTIN_EMOJIS[unicode_name]
            continue
        
        yield Emoji._create_partial(
            int(custom_emoji_id),
            custom_name,
            (True if custom_animated else False),
        )
        continue


def parse_custom_emojis(text):
    """
    Parses out every custom emoji from the given text.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Returns
    -------
    emojis : `set` of ``Emoji``
    """
    if text is None:
        return set()
    
    return {*_iter_parse_custom_emojis(text)}


def parse_all_emojis(text):
    """
    Parses out every emoji from the given text.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Returns
    -------
    emojis : `set` of ``Emoji``
    """
    if text is None:
        return set()
    return {*_iter_parse_all_emojis(text)}
    


def _parse_emojis_ordered(text, parser):
    """
    Parses emojis of the given `text` with the given `parser`.
    Returns them ordered based on their appearance in the text.
    
    
    Parameters
    ----------
    text : `None`, `str`
        The text to parse.
    parser : `GeneratorFunction`
        The parser to use.
    
    Returns
    -------
    emojis_ordered : `list` of ``Emoji``
        Excludes duplicates.
    """
    emojis_ordered = []
    if (text is not None):
        emojis_unique = set()
        
        for emoji in parser(text):
            if emoji not in emojis_unique:
                emojis_ordered.append(emoji)
                emojis_unique.add(emoji)
    
    return emojis_ordered


def parse_custom_emojis_ordered(text):
    """
    Parses out every custom emoji from the given text. Returns them ordered based on their appearance in the text.
    
    Parameters
    ----------
    text : `None`, `str`
        The text to parse.
    
    Returns
    -------
    emojis_ordered : `list` of ``Emoji``
        Excludes duplicates.
    """
    return _parse_emojis_ordered(text, _iter_parse_custom_emojis)


def parse_all_emojis_ordered(text):
    """
    Parses out every emoji from the given text. Returns them ordered based on their appearance in the text.
    
    Parameters
    ----------
    text : `None`, `str`
        The text to parse.
    
    Returns
    -------
    emojis_ordered : `list` of ``Emoji``
        Excludes duplicates.
    """
    return _parse_emojis_ordered(text, _iter_parse_all_emojis)


def parse_reaction(text):
    """
    Parses out an emoji from the given reaction string.
    
    Parameters
    ----------
    text : `str`
        Reaction string.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    try:
        emoji = UNICODE_TO_EMOJI[text]
    except KeyError:
        parsed = REACTION_RP.fullmatch(text)
        if parsed is None:
            emoji = None
        else:
            name, emoji_id = parsed.groups()
            emoji_id = int(emoji_id)
            emoji = Emoji._create_partial(emoji_id, name, False)
    
    return emoji



__all__ = ('ApplicationCommandTargetType', 'APPLICATION_COMMAND_CONTEXT_TARGET_TYPES')

from ...bases import Preinstance as P, PreinstancedBase


class ApplicationCommandTargetType(PreinstancedBase):
    """
    Represents an application command's target.
    
    Attributes
    ----------
    name : `str`
        The name of the application command target.
    value : `int`
        The identifier value the application command target.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationCommandTargetType``) items
        Stores the predefined ``ApplicationCommandTargetType``-s. These can be accessed with their `value` as
        key.
    VALUE_TYPE : `type` = `int`
        The application command targets' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the application command targets.
    
    Every predefined application command target can be accessed as class attribute as well:
    
    +-----------------------+-------------------+-------+
    | Class attribute name  | Name              | Value |
    +=======================+===================+=======+
    | none                  | none              | 0     |
    +-----------------------+-------------------+-------+
    | chat                  | chat              | 1     |
    +-----------------------+-------------------+-------+
    | user                  | user              | 2     |
    +-----------------------+-------------------+-------+
    | message               | message           | 3     |
    +-----------------------+-------------------+-------+
    | channel               | channel           | 4     |
    +-----------------------+-------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none',)
    chat = P(1, 'chat',)
    user = P(2, 'user',)
    message = P(3, 'message',)
    channel = P(4, 'channel')


APPLICATION_COMMAND_CONTEXT_TARGET_TYPES = frozenset((
    ApplicationCommandTargetType.user,
    ApplicationCommandTargetType.message,
    ApplicationCommandTargetType.channel
))



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_optional_putter_factory
from ...field_validators import nullable_string_validator_factory

from ..shared_fields import validate_components as validate_any_components

from ..component import Component, ComponentType

from .constants import TITLE_LENGTH_MIN, TITLE_LENGTH_MAX

# components

def validate_components(components):
    """
    Interaction form specific component validator.
    
    Parameters
    ----------
    components : `None`, `iterable` of ``Component``
        The components to validate.
    
    Returns
    -------
    components : `None`, `tuple` of ``Component``
    """
    components = validate_any_components(components)
    if (components is not None):
        return tuple(
            component if (component.type is ComponentType.row)
            else Component(ComponentType.row, components = [component])
            for component in components
        )

# title

parse_title = nullable_string_parser_factory('title')
put_title_into = nullable_string_optional_putter_factory('title')
validate_title = nullable_string_validator_factory('title', TITLE_LENGTH_MIN, TITLE_LENGTH_MAX)



__all__ = ()

from .fields.allow import validate_allow, put_allow_into
from .fields.deny import validate_deny, put_deny_into
from .fields.target import validate_target, put_target_into
from .fields.target_id import validate_target_id, put_target_id_into
from .fields.target_type import validate_target_type, put_target_type_into


PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS = {
    'allow': (validate_allow, put_allow_into),
    'deny': (validate_deny, put_deny_into),
}


PERMISSION_OVERWRITE_FIELD_CONVERTERS = {
    **PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS,
    'target': (validate_target, put_target_into),
    'target_id': (validate_target_id, put_target_id_into),
    'target_type': (validate_target_type, put_target_type_into),
}



from .emoji import *
from .parsing import *
from .reaction import *
from .reaction_events import *
from .unicode import *


__all__ = (
    *emoji.__all__,
    *parsing.__all__,
    *reaction.__all__,
    *reaction_events.__all__,
    *unicode.__all__,
)

# Deprecations

from ...utils.module_deprecation import deprecated_import

deprecated_import(ReactionMapping, 'reaction_mapping')
deprecated_import(ReactionMappingLine, 'reaction_mapping_line')
